import unittest
import requests_mock
from test.utils import assert_matching_tables

from parsons import Table, Zoom

API_KEY = 'fake_api_key'
API_SECRET = 'fake_api_secret'
ZOOM_URI = 'https://api.zoom.us/v2/'


class TestZoom(unittest.TestCase):

    def setUp(self):
        self.zoom = Zoom(API_KEY, API_SECRET)

    @requests_mock.Mocker()
    def test_get_users(self, m):
        user_json = {
            'page_count': 1,
            'page_number': 1,
            'page_size': 30,
            'total_records': 1,
            'users': [{
                'id': 'C5A2nRWwTMm_hXyJb1JXMh',
                'first_name': 'Bob',
                'last_name': 'McBob',
                'email': 'bob@bob.com',
                'type': 2,
                'pmi': 8374523641,
                'timezone': 'America/New_York',
                'verified': 1,
                'dept': '',
                'created_at': '2017-10-06T15:22:34Z',
                'last_login_time': '2020-05-06T16:50:45Z',
                'last_client_version': '',
                'language': '',
                'phone_number': '',
                'status': 'active'}]}

        tbl = Table([{'id': 'C5A2nRWwTMm_hXyJb1JXMh',
                      'first_name': 'Bob',
                      'last_name': 'McBob',
                      'email': 'bob@bob.com',
                      'type': 2,
                      'pmi': 8374523641,
                      'timezone': 'America/New_York',
                      'verified': 1,
                      'dept': '',
                      'created_at': '2017-10-06T15:22:34Z',
                      'last_login_time': '2020-05-06T16:50:45Z',
                      'last_client_version': '',
                      'language': '',
                      'phone_number': '',
                      'status': 'active'}])

        m.get(ZOOM_URI + 'users', json=user_json)
        assert_matching_tables(self.zoom.get_users(), tbl)

    @requests_mock.Mocker()
    def test_get_meeting_participants(self, m):
        participants = {
            'page_count': 1,
            'page_size': 30,
            'total_records': 4,
            'next_page_token': '',
            'participants': [{
                'id': '',
                'user_id': '16778240',
                'name': 'Barack Obama',
                'user_email': '',
                'join_time': '2020-04-24T21:00:26Z',
                'leave_time': '2020-04-24T22:24:38Z',
                'duration': 5052,
                'attentiveness_score': ''
            }, {
                'id': '',
                'user_id': '16779264',
                'name': '',
                'user_email': '',
                'join_time': '2020-04-24T21:00:45Z',
                'leave_time': '2020-04-24T22:24:38Z',
                'duration': 5033,
                'attentiveness_score': ''
            }]}

        tbl = Table([{
            'id': '',
            'user_id': '16778240',
            'name': 'Barack Obama',
            'user_email': '',
            'join_time': '2020-04-24T21:00:26Z',
            'leave_time': '2020-04-24T22:24:38Z',
            'duration': 5052,
            'attentiveness_score': ''
        }, {
            'id': '',
            'user_id': '16779264',
            'name': '',
            'user_email': '',
            'join_time': '2020-04-24T21:00:45Z',
            'leave_time': '2020-04-24T22:24:38Z',
            'duration': 5033,
            'attentiveness_score': ''}])

        m.get(ZOOM_URI + 'report/meetings/123/participants', json=participants)
        assert_matching_tables(self.zoom.get_past_meeting_participants(123), tbl)

    @requests_mock.Mocker()
    def test_get_meeting_registrants(self, m):
        registrants = {
            'page_count': 1,
            'page_size': 30,
            'total_records': 4,
            'next_page_token': '',
            'registrants': [{
                'id': '',
                'user_id': '16778240',
                'name': 'Barack Obama',
                'user_email': '',
                'purchasing_time_frame': 'Within a month',
                'role_in_purchase_process': 'Not involved'
            }, {
                'id': '',
                'user_id': '16779264',
                'name': '',
                'user_email': '',
                'purchasing_time_frame': 'Within a month',
                'role_in_purchase_process': 'Not involved'
            }]}

        tbl = Table([
            {
                'id': '',
                'user_id': '16778240',
                'name': 'Barack Obama',
                'user_email': '',
                'purchasing_time_frame': 'Within a month',
                'role_in_purchase_process': 'Not involved'
            },
            {
                'id': '',
                'user_id': '16779264',
                'name': '',
                'user_email': '',
                'purchasing_time_frame': 'Within a month',
                'role_in_purchase_process': 'Not involved'
            }
        ])

        m.get(ZOOM_URI + 'meetings/123/registrants', json=registrants)
        assert_matching_tables(self.zoom.get_meeting_registrants(123), tbl)

    @requests_mock.Mocker()
    def test_get_user_webinars(self, m):
        webinars = {
            'page_count': 1,
            'page_size': 30,
            'total_records': 4,
            'next_page_token': '',
            'webinars': [{
                "uuid": "dsghfkhaewfds",
                "id": '',
                "host_id": "24654130000000",
                "topic": "My Webinar",
                "agenda": "Learn more about Zoom APIs",
                "type": "5",
                "duration": "60",
                "start_time": "2019-09-24T22:00:00Z",
                "timezone": "America/Los_Angeles",
                "created_at": "2019-08-30T22:00:00Z",
                "join_url": "https://zoom.us/0001000/awesomewebinar"
            }, {
                "uuid": "dhf8as7dhf",
                "id": '',
                "host_id": "24654130000345",
                "topic": "My Webinar",
                "agenda": "Learn more about Zoom APIs",
                "type": "5",
                "duration": "60",
                "start_time": "2019-09-24T22:00:00Z",
                "timezone": "America/Los_Angeles",
                "created_at": "2019-08-30T22:00:00Z",
                "join_url": "https://zoom.us/0001000/awesomewebinar"
            }]}

        tbl = Table([
            {
                "uuid": "dsghfkhaewfds",
                "id": '',
                "host_id": "24654130000000",
                "topic": "My Webinar",
                "agenda": "Learn more about Zoom APIs",
                "type": "5",
                "duration": "60",
                "start_time": "2019-09-24T22:00:00Z",
                "timezone": "America/Los_Angeles",
                "created_at": "2019-08-30T22:00:00Z",
                "join_url": "https://zoom.us/0001000/awesomewebinar"
            },
            {
                "uuid": "dhf8as7dhf",
                "id": '',
                "host_id": "24654130000345",
                "topic": "My Webinar",
                "agenda": "Learn more about Zoom APIs",
                "type": "5",
                "duration": "60",
                "start_time": "2019-09-24T22:00:00Z",
                "timezone": "America/Los_Angeles",
                "created_at": "2019-08-30T22:00:00Z",
                "join_url": "https://zoom.us/0001000/awesomewebinar"
            }
        ])

        m.get(ZOOM_URI + 'users/123/webinars', json=webinars)
        assert_matching_tables(self.zoom.get_user_webinars(123), tbl)

    @requests_mock.Mocker()
    def test_get_past_webinar_participants(self, m):
        participants = {
            'page_count': 1,
            'page_size': 30,
            'total_records': 4,
            'next_page_token': '',
            'participants': [{
                "id": "",
                "user_id": "sdfjkldsf87987",
                "name": "Barack",
                "user_email": "riya@sdfjkldsf87987.fdjfhdf",
                "join_time": "2019-02-01T12:34:12.660Z",
                "leave_time": "2019-03-01T12:34:12.660Z",
                "duration": "20"
            }, {
                "id": "",
                "user_id": "sdfjkldsfdfgdfg",
                "name": "Joe",
                "user_email": "riya@sdfjkldsf87987.fdjfhdf",
                "join_time": "2019-02-01T12:34:12.660Z",
                "leave_time": "2019-03-01T12:34:12.660Z",
                "duration": "20"
            }]}

        tbl = Table([
            {
                "id": "",
                "user_id": "sdfjkldsf87987",
                "name": "Barack",
                "user_email": "riya@sdfjkldsf87987.fdjfhdf",
                "join_time": "2019-02-01T12:34:12.660Z",
                "leave_time": "2019-03-01T12:34:12.660Z",
                "duration": "20"
            },
            {
                "id": "",
                "user_id": "sdfjkldsfdfgdfg",
                "name": "Joe",
                "user_email": "riya@sdfjkldsf87987.fdjfhdf",
                "join_time": "2019-02-01T12:34:12.660Z",
                "leave_time": "2019-03-01T12:34:12.660Z",
                "duration": "20"
            }
        ])

        m.get(ZOOM_URI + 'report/webinars/123/participants', json=participants)
        assert_matching_tables(self.zoom.get_past_webinar_participants(123), tbl)

    @requests_mock.Mocker()
    def test_get_webinar_registrants(self, m):
        registrants = {
            'page_count': 1,
            'page_size': 30,
            'total_records': 4,
            'next_page_token': '',
            'registrants': [{
                "id": "",
                "email": "barack@obama.com",
                "first_name": "Barack",
                "last_name": "Obama",
                "address": "dsfhkdjsfh st",
                "city": "jackson heights",
                "country": "US",
                "zip": "11371",
                "state": "NY",
                "phone": "00000000",
                "industry": "Food",
                "org": "Cooking Org",
                "job_title": "Chef",
                "purchasing_time_frame": "1-3 months",
                "role_in_purchase_process": "Influencer",
                "no_of_employees": "10",
                "comments": "Looking forward to the Webinar",
                "custom_questions": [
                    {
                      "title": "What do you hope to learn from this Webinar?",
                      "value": "Look forward to learning how you come up with recipes and services"
                    }
                ],
                "status": "approved",
                "create_time": "2019-02-26T23:01:16.899Z",
                "join_url": "https://zoom.us/webinar/mywebinariscool"
            }, {
                "id": "",
                "email": "joe@biden.com",
                "first_name": "Joe",
                "last_name": "Biden",
                "address": "dsfhkdjsfh st",
                "city": "jackson heights",
                "country": "US",
                "zip": "11371",
                "state": "NY",
                "phone": "00000000",
                "industry": "Food",
                "org": "Cooking Org",
                "job_title": "Chef",
                "purchasing_time_frame": "1-3 months",
                "role_in_purchase_process": "Influencer",
                "no_of_employees": "10",
                "comments": "Looking forward to the Webinar",
                "custom_questions": [
                    {
                      "title": "What do you hope to learn from this Webinar?",
                      "value": "Look forward to learning how you come up with recipes and services"
                    }
                ],
                "status": "approved",
                "create_time": "2019-02-26T23:01:16.899Z",
                "join_url": "https://zoom.us/webinar/mywebinariscool"
            }]}

        tbl = Table([
            {
                "id": "",
                "email": "barack@obama.com",
                "first_name": "Barack",
                "last_name": "Obama",
                "address": "dsfhkdjsfh st",
                "city": "jackson heights",
                "country": "US",
                "zip": "11371",
                "state": "NY",
                "phone": "00000000",
                "industry": "Food",
                "org": "Cooking Org",
                "job_title": "Chef",
                "purchasing_time_frame": "1-3 months",
                "role_in_purchase_process": "Influencer",
                "no_of_employees": "10",
                "comments": "Looking forward to the Webinar",
                "custom_questions": [
                    {
                      "title": "What do you hope to learn from this Webinar?",
                      "value": "Look forward to learning how you come up with recipes and services"
                    }
                ],
                "status": "approved",
                "create_time": "2019-02-26T23:01:16.899Z",
                "join_url": "https://zoom.us/webinar/mywebinariscool"
            },
            {
                "id": "",
                "email": "joe@biden.com",
                "first_name": "Joe",
                "last_name": "Biden",
                "address": "dsfhkdjsfh st",
                "city": "jackson heights",
                "country": "US",
                "zip": "11371",
                "state": "NY",
                "phone": "00000000",
                "industry": "Food",
                "org": "Cooking Org",
                "job_title": "Chef",
                "purchasing_time_frame": "1-3 months",
                "role_in_purchase_process": "Influencer",
                "no_of_employees": "10",
                "comments": "Looking forward to the Webinar",
                "custom_questions": [
                    {
                      "title": "What do you hope to learn from this Webinar?",
                      "value": "Look forward to learning how you come up with recipes and services"
                    }
                ],
                "status": "approved",
                "create_time": "2019-02-26T23:01:16.899Z",
                "join_url": "https://zoom.us/webinar/mywebinariscool"
            }
        ])

        m.get(ZOOM_URI + 'webinars/123/registrants', json=registrants)
        assert_matching_tables(self.zoom.get_webinar_registrants(123), tbl)



import json
import logging

from parsons.utilities import check_env
from parsons.utilities.api_connector import APIConnector

logger = logging.getLogger(__name__)

URI = 'https://app.periscopedata.com/api/v1/'


class Sisense(object):
    """
    Instantiate the Sisense class.

    `Args:`
        site_name: str
            The name of the site. Not required if the ``SISENSE_SITE_NAME``
            environmental variable is set.
        api_key: str
            The Sisense API Key. Not required if the ``SISENSE_API_KEY``
            environmental variable is set.
    `Returns:`
        Sisense class
    """

    def __init__(self, site_name=None, api_key=None):
        self.site_name = check_env.check('SISENSE_SITE_NAME', site_name)
        self.api_key = check_env.check('SISENSE_API_KEY', api_key)
        self.uri = URI
        self.api = self._api()

    def _api(self):
        headers = {'HTTP-X-PARTNER-AUTH': self.site_name + ":" + self.api_key}
        return APIConnector(uri=self.uri, headers=headers)

    def publish_shared_dashboard(self, dashboard_id, chart_id=None, **kwargs):
        """
        This method publishes a dashboard or chart using the provided arguments.
        For available options, see the `API documentation <https://dtdocs.sisense.com/article/embed-api-options>`_. # noqa

        `Args:`
            dashboard_id: str or int
                The ID of the dashboard (required).
            chart_id: str or int
                The ID of the chart. Only required for publishing individual charts.
            **kwargs:
                Optional arguments.
        `Returns:`
            Response (dict containing the URL) or an error
        """
        payload = {'dashboard': dashboard_id, 'chart': chart_id, **kwargs}
        return self.api.post_request('shared_dashboard/create', data=json.dumps(payload))

    def list_shared_dashboards(self, dashboard_id):
        """
        List all shares of a given dashboard.

        `Args:`
            dashboard_id: str or int
                The ID the dashboard (required).
        `Returns:`
            Response or an error
        """
        payload = {'dashboard': dashboard_id}
        return self.api.post_request('shared_dashboard/list', data=json.dumps(payload))

    def delete_shared_dashboard(self, token):
        """
        To delete a shared dashboard you must provide the token for the shared dashboard.
        The token is the last part of the shared dashboard URL. i.e. if the shared URL is:

        https://app.periscopedata.com/shared/9dda9dda-9dda-9dda-9dda-9dda9dda9dda

        The token is '9dda9dda-9dda-9dda-9dda-9dda9dda9dda'.

        `Args:`
            token: str or int
                The token of the shared dashboard (required).
        `Returns:`
            Response or an error
        """
        payload = {'token': token}
        return self.api.post_request('shared_dashboard/delete', data=json.dumps(payload))



import logging

logger = logging.getLogger(__name__)


class Events:
    """A class for interacting with PDI events via PDIs API"""

    def __init__(self):
        self.events_url = self.base_url + '/events'
        self.calendars_url = self.base_url + '/calendars'
        self.eventactivities_url = self.base_url + '/eventActivities'
        self.activites_url = self.base_url + '/activities'
        self.activityassignment_url = self.base_url + '/eventActivityAssignements'

        super().__init__()

    def get_events(self, first_event_date: str, last_event_date: str, limit=None):
        """Get a table of PDI events in a given time frame

        `Args:`
            first_event_date: str
                First date in the timeframe from which you want events formatted at 'yyy-MM-dd'
            last_event_date: str
                Last date in the timeframe from which you want events formatted at 'yyy-MM-dd'
            limit: int
                The max number of events to return

        `Returns:`
            parsons.Table
                A Parsons table containing all requested events data.
        """

        params = {
            'startDate': first_event_date,
            'endDate': last_event_date,
        }

        return self._request(self.events_url, args=params, limit=limit)

    def get_event_invitations(self, event_id: str, expand=True, limit=None):
        """Get a table of PDI event invitations for a specified event

        `Args:`
            event_id: str
                ID of event for which you want invitations
            expand: bool
                If True returns columns for contact (and all contact info) and event)

        `Returns:`
            parsons.Table
                A Parsons table containing all requested event invitation data.
        """

        params = {'expand': expand}

        return self._request(f'{self.events_url}/{event_id}/invitations', args=params, limit=limit)

    def create_event(self, calendar_id: str, location_id: str, event_name: str, start_datetime: str,
                     end_datetime: str, description=None, all_day=False, recurrencetype=None,
                     recurrence_end_datetime=None, host_phone=None, host_email=None, website=None):
        """Create event in a specified calendar

        `Args:`
            calendar_id: str
                The calendar in which you'd like to create an event
            location_id: str
                The unique ID of the PDI location this event took place/is to take place at
            event_name: str
                The name of your event
            description: str
                A short description for your event
            start_datetime: str
                The start datetime of the event in UTC timezone formatted as
                yyyy-MM-ddThh:mm:ss.fffZ
            end_datetime: str
                The end date formatted like start_datetime
            is_all_day: bool
                set to True if event is an all day event. Defaults to False
            recurrencetype: str
                Either 'daily', 'weekly', or 'monthly'. Defaults to None
            recurrence_end_datetime: str
                The end time of the last recurrence of the event formatted as
                yyyy-MM-ddThh:mm:ss.fffZ
            host_phone: str
                An optional contact phone number for the host. Defaults to None
            host_email: str
                An optional contact email for the host. Defaults to None
            website: str
                An optional website for the event. Defualts to None

        `Returns:`
            dict
                Response from PDI in dictionary object

            """

        payload = {
          "locationId": location_id,
          "recurrenceType": recurrencetype,
          "name": event_name,
          "description": description,
          "startDateTimeUtc": start_datetime,
          "endDateTimeUtc": end_datetime,
          "isAllDay": str(all_day).lower(),
          "recurrenceEndDateTimeUtc": recurrence_end_datetime,
          "phone": host_phone,
          "contactEmail": host_email,
          "website": website
        }

        response = self._request(self.calendars_url + f'/{calendar_id}' + '/events',
                                 req_type='POST', post_data=payload)
        event_id = response['id']
        logger.info(f'Created event {event_name} (id: {event_id})')

        return response

    def create_event_with_activity(self, calendar_id: str, location_id: str, activity_id: str,
                                   event_name: str, activity_name: str, start_datetime: str,
                                   end_datetime: str, description=None, all_day=False,
                                   recurrencetype=None, recurrence_end_datetime=None,
                                   host_phone=None, host_email=None, website=None,
                                   signup_goal=None):
        """Create event in a specified calendar with an associated activity. The activty will
            be assigned the same start, end time, and recurrance settings as the event.

                `Args:`
                    calendar_id: str
                        The unique ID of the calendar in which you'd like to create an event
                    location_id: str
                        The unique ID of the PDI location whek this event took place/is to take
                        place
                    activity_id:
                        The unique ID of the activity type you'd like to add to the event
                    event_name: str
                        The name of your event
                    activity_name: str
                        The name of your activity. e.g. 'Pictionary!'
                    description: str
                        A short description for your event
                    start_datetime: str
                        The start datetime of the event in UTC timezone formatted as
                        yyyy-MM-ddThh:mm:ss.fffZ
                    end_datetime: str
                        The end date formatted like start_datetime
                    is_all_day = bool
                        set to True if event is an all day event. Defaults to False
                    recurrencetype: str
                        Either 'daily', 'weekly', or 'monthly'. Defaults to None
                    recurrence_end_datetime: str
                        The end time of the last recurrence of the event formatted as
                        yyyy-MM-ddThh:mm:ss.fffZ
                    host_phone: str
                        An optional contact phone number for the host. Defaults to None
                    host_email: str
                        An optional contact email for the host. Defaults to None
                    website: str
                        An optional website for the event. Defualts to None
                    signup_goal: int
                        The goal of how many people you want to complete the activity
                `Returns:`
                    dict
                        Response from PDI in dictionary object
                    """
        event_data = self.create_event(calendar_id, location_id, event_name, start_datetime,
                                       end_datetime, description, all_day, recurrencetype,
                                       recurrence_end_datetime, host_phone, host_email, website)
        event_id = event_data['id']
        logger.info(f'Created event {event_name} (id: {event_id})')

        event_activity_payload = {
            "CalendarId": calendar_id,
            "EventId": event_id,
            "ActivityId": activity_id,
            "LocationId": location_id,
            "RecurrenceType": recurrencetype,
            "Name": activity_name,
            "Description": None,
            "StartDateTimeUtc": start_datetime,
            "EndDateTimeUtc": end_datetime,
            "CountGoal": signup_goal,
            "RecurrenceEndDateTimeUtc": recurrence_end_datetime
        }

        response = self._request(self.eventactivities_url, req_type='POST',
                                 post_data=event_activity_payload)
        logger.info(f'Created activity {activity_name} for event {event_name} (id: {event_id})')

        return response

    def create_event_activity(self, calendar_id: str, event_id: str, activity_id: str,
                              location_id: str, activity_name: str, start_datetime: str,
                              end_datetime: str, description=None, recurrencetype=None,
                              recurrence_end_datetime=None, signup_goal=None):
        """Create event in a specified calendar with an associated activity

                `Args:`
                    calendar_id: str
                        The unique ID of the calendar in which you'd like to create an event
                    event_id: str
                        The unique ID of the event this activity is to be associated with
                    activity_id:
                        The unique ID of the activity type you'd like to add to the event
                    location_id: str
                        The unique ID of the PDI location where this event took place/is to take
                        place
                    activity_name: str
                        The name of your activity. e.g. 'Pictionary!'
                    description: str
                        A short description for your event activity
                    start_datetime: str
                        The start datetime of the event in UTC timezone formatted as
                        yyyy-MM-ddThh:mm:ss.fffZ
                    end_datetime: str
                        The end date formatted like start_datetime
                    recurrencetype: str
                        Either 'daily', 'weekly', or 'monthly'. Defaults to None
                    recurrence_end_datetime: str
                        The end time of the last recurrence of the event formatted as
                        yyyy-MM-ddThh:mm:ss.fffZ
                    signup_goal: int
                        The goal of how many people you want to complete the activity


                `Returns:`
                    dict
                        Response from PDI in dictionary object
                    """

        event_activity_payload = {
            "CalendarId": calendar_id,
            "EventId": event_id,
            "ActivityId": activity_id,
            "LocationId": location_id,
            "RecurrenceType": recurrencetype,
            "Name": activity_name,
            "Description": description,
            "StartDateTimeUtc": start_datetime,
            "EndDateTimeUtc": end_datetime,
            "CountGoal": signup_goal,
            "RecurrenceEndDateTimeUtc": recurrence_end_datetime
        }

        response = self._request(self.eventactivities_url, req_type='POST',
                                 post_data=event_activity_payload)
        logger.info(f'Created activity {activity_name} for event {event_id})')

        return response

    def create_invitation(self, event_id: str, contact_id: str, status: str, attended: bool,
                          confirmed=False, specific_occurrence_start=None):
        """Create a PDI event invitation indicating a contact has been registered for an event
            `Args:`
                event_id: str
                    The ID of the event to write the RSVP to
                contact_id: str
                    The ID of the contact to which the invitation belongs
                status: str
                    Options are: "Yes", "No", "Maybe", "Scheduled", "Invited", "Cancelled",
                    "No-Show", "Completed", and ""
                attended: boolean
                    Indicates whether contact attended event
                confirmed: boolean
                    Indicates whether invitation confirmed they will attend the event. Defaults to
                    False
                specific_occurrence_start: str
                    If invitation is for a specific occurrence of a recurring event, then the start
                    datetime of the event in UTC formatted as yyyy-MM-ddTHH:mm:ss.fffZ
            `Returns:`
                dict
                    Response from PDI in dictionary object
        """

        event_invitation_payload = {
            "contactId": contact_id,
            "rsvpStatus": status,
            "isConfirmed": confirmed,
            "attended": attended
        }

        if specific_occurrence_start:
            event_invitation_payload["specificOcurrenceStartUtc"] = specific_occurrence_start

        response = self._request(self.events_url + f'/{event_id}/invitations',
                                 req_type='POST', post_data=event_invitation_payload)
        return response

    def update_invitation(self, invitation_id: str, event_id: str, contact_id: str, status=None,
                          attended=None, confirmed=None, specific_occurrence_start=None):
        """Modify a PDI event invitation
            `Args:`
                invitation_id: str
                    The ID of the event invitation
                event_id: str
                    The ID of the event that corresponds to the invitation
                contact_id: str
                    The ID of the contact to which the invitation belongs
                status: str
                    Options are: "Yes", "No", "Maybe", "Scheduled", "Invited", "Cancelled",
                    "No-Show", "Completed", and ""
                attended: boolean
                    Indicates whether contact attended event
                confirmed: boolean
                    Indicates whether invitation confirmed they will attend the event
                specific_occurrence_start: str
                    If invitation is for a specific occurrence of a recurring event, then the start
                    datetime of the event in UTC formatted as yyyy-MM-ddTHH:mm:ss.fffZ
            `Returns:`
                dict
                    Response from PDI in dictionary object
        """

        event_invitation_payload = {
            "contactId": contact_id
        }

        if status:
            event_invitation_payload['rsvpStatus'] = status
        if confirmed is not None:
            event_invitation_payload['isConfirmed'] = confirmed
        if attended is not None:
            event_invitation_payload['attended'] = attended
        if specific_occurrence_start:
            event_invitation_payload["specificOcurrenceStartUtc"] = specific_occurrence_start

        response = self._request(self.events_url + f'/{event_id}/invitations/{invitation_id}',
                                 req_type='PUT', post_data=event_invitation_payload)
        return response

    def create_activity_assignment(self, eventactivityid: str, contact_id: str, status: str,
                                   completed: bool, confirmed=False,
                                   specific_occurrence_start=None):
        """Create an activity assignement
            `Args:`
                eventactivityid: str
                    The ID of the specific event activity you'd like to assign a contact
                contact_id: str
                    The ID of the contact to which the assignment belongs
                status: str
                    Options are: "Yes", "No", "Maybe", "Scheduled", "Invited", "Cancelled",
                    "No-Show", "Completed", and ""
                completed: boolean
                    Indicates whether contact attended event
                confirmed: boolean
                    Indicates whether invitation confirmed they will attend the event
                specific_occurrence_start: str
                    If invitation is for a specific occurrence of a recurring event, then the start
                    datetime of the event in UTC formatted as yyyy-MM-ddTHH:mm:ss.fffZ
            `Returns:`
                dict
                    Response from PDI in dictionary object
        """

        assignment_payload = {
            "rsvpStatus": status,
            "isConfirmed": confirmed,
            "isShiftWorked": completed,
            "contactId": contact_id,
            "eventActivityId": eventactivityid
        }

        if specific_occurrence_start:
            assignment_payload["specificOcurrenceStartUtc"] = specific_occurrence_start

        response = self._request(self.activityassignment_url, req_type='POST',
                                 post_data=assignment_payload)

        return response

    def update_activity_assignment(self, activityassignementid: str, eventactivityid: str,
                                   contact_id: str, status=None, completed=None, confirmed=None,
                                   specific_occurrence_start=None):
        """Create an activity assignement
            `Args:`
                activityassignementid: str
                    Id of the specific event activity assignement you want to modify
                eventactivityid: str
                    The ID of the specific event activity you'd like to assign a contact
                contact_id: str
                    The ID of the contact to which the assignment belongs
                status: str
                    Options are: "Yes", "No", "Maybe", "Scheduled", "Invited", "Cancelled",
                    "No-Show", "Completed", and ""
                completed: boolean
                    Indicates whether contact attended event
                confirmed: boolean
                    Indicates whether invitation confirmed they will attend the event
                specific_occurrence_start: str
                    If invitation is for a specific occurrence of a recurring event, then the start
                    datetime of the event in UTC formatted as yyyy-MM-ddTHH:mm:ss.fffZ
            `Returns:`
                dict
                    Response from PDI in dictionary object
        """

        assignment_payload = {
            "contactId": contact_id,
            "eventActivityId": eventactivityid
        }

        if status:
            assignment_payload['rsvpStatus'] = status
        if confirmed is not None:
            assignment_payload['isConfirmed'] = confirmed
        if completed is not None:
            assignment_payload['isShiftWorked'] = completed
        if specific_occurrence_start:
            assignment_payload["specificOcurrenceStartUtc"] = specific_occurrence_start

        response = self._request(self.activityassignment_url + f'/{activityassignementid}',
                                 req_type='PUT', post_data=assignment_payload)

        return response



# Dictionaries used to map column names to valid argument names

LEAD_COLUMN_MAP = {'first_name': ['first', 'fn'],
                   'last_name': ['last', 'ln', 'lastname'],
                   'phone_number': ['phone', 'cell', 'cellphone'],
                   'email': ['emailaddress'],
                   'follow_up': ['follow']
                   }



from parsons.databases.postgres.postgres_core import PostgresCore
from parsons.databases.table import BaseTable
from parsons.databases.alchemy import Alchemy
import logging
import os


logger = logging.getLogger(__name__)


class Postgres(PostgresCore, Alchemy):
    """
    A Postgres class to connect to database. Credentials can be passed from a ``.pgpass`` file
    stored in your home directory or with environmental variables.

    Args:
        username: str
            Required if env variable ``PGUSER`` not populated
        password: str
            Required if env variable ``PGPASSWORD`` not populated
        host: str
            Required if env variable ``PGHOST`` not populated
        db: str
            Required if env variable ``PGDATABASE`` not populated
        port: int
            Required if env variable ``PGPORT`` not populated.
        timeout: int
            Seconds to timeout if connection not established.
    """

    def __init__(self, username=None, password=None, host=None, db=None, port=5432, timeout=10):
        super().__init__()

        self.username = username or os.environ.get('PGUSER')
        self.password = password or os.environ.get('PGPASSWORD')
        self.host = host or os.environ.get('PGHOST')
        self.db = db or os.environ.get('PGDATABASE')
        self.port = port or os.environ.get('PGPORT')

        # Check if there is a pgpass file. Psycopg2 will search for this file first when
        # creating a connection.
        pgpass = os.path.isfile(os.path.expanduser('~/.pgpass'))

        if not any([self.username, self.password, self.host, self.db]) and not pgpass:
            raise ValueError('Connection arguments missing. Please pass as a pgpass file, kwargs',
                             'or env variables.')

        self.timeout = timeout
        self.dialect = 'postgres'

    def copy(self, tbl, table_name, if_exists='fail', strict_length=False):
        """
        Copy a :ref:`parsons-table` to Postgres.

        `Args:`
            tbl: parsons.Table
                A Parsons table object
            table_name: str
                The destination schema and table (e.g. ``my_schema.my_table``)
            if_exists: str
                If the table already exists, either ``fail``, ``append``, ``drop``
                or ``truncate`` the table.
            strict_length: bool
                If the database table needs to be created, strict_length determines whether
                the created table's column sizes will be sized to exactly fit the current data,
                or if their size will be rounded up to account for future values being larger
                then the current dataset
        """

        with self.connection() as connection:

            # Auto-generate table
            if self._create_table_precheck(connection, table_name, if_exists):

                # Create the table
                # To Do: Pass in the advanced configuration parameters.
                sql = self.create_statement(tbl, table_name, strict_length=strict_length)

                self.query_with_connection(sql, connection, commit=False)
                logger.info(f'{table_name} created.')

            sql = f"COPY {table_name} FROM STDIN CSV HEADER;"

            with self.cursor(connection) as cursor:
                cursor.copy_expert(sql, open(tbl.to_csv(), "r"))
                logger.info(f'{tbl.num_rows} rows copied to {table_name}.')

    def table(self, table_name):
        # Return a Postgres table object

        return PostgresTable(self, table_name)


class PostgresTable(BaseTable):
    # Postgres table object.

    pass



from oauthlib.oauth2 import BackendApplicationClient
from requests_oauthlib import OAuth2Session
from parsons.utilities.api_connector import APIConnector
import urllib.parse


class OAuth2APIConnector(APIConnector):
    """
    The OAuth2API Connector is a low level class for authenticated API requests using OAuth2.
    It extends APIConnector by wrapping the request methods in a server-side OAuth2 client
    and otherwise provides the same interface as APIConnector.

    `Args:`
        uri: str
            The base uri for the api. Must include a trailing '/' (e.g. ``http://myapi.com/v1/``)
        headers: dict
            The request headers
        auth: dict
            The request authorization parameters
        pagination_key: str
            The name of the key in the response json where the pagination url is
            located. Required for pagination.
        data_key: str
            The name of the key in the response json where the data is contained. Required
            if the data is nested in the response json
        client_id: str
            The client id for acquiring and exchanging tokens from the OAuth2 application
        client_secret: str
            The client secret for acquiring and exchanging tokens  from the OAuth2 application
        token_url: str
            The URL for acquiring new tokens from the OAuth2 Application
        auto_refresh_url: str
            If provided, the URL for refreshing tokens from the OAuth2 Application
    `Returns`:
        OAuthAPIConnector class
    """

    def __init__(
        self, uri, headers=None, auth=None, pagination_key=None, data_key=None,
        client_id=None, client_secret=None, token_url=None, auto_refresh_url=None
    ):
        super().__init__(
            uri, headers=headers, auth=auth, pagination_key=pagination_key, data_key=data_key
        )

        client = BackendApplicationClient(client_id=client_id)
        oauth = OAuth2Session(client=client)
        self.token = oauth.fetch_token(token_url=token_url,
                                       client_id=client_id, client_secret=client_secret)
        self.client = OAuth2Session(
            client_id, token=self.token, auto_refresh_url=auto_refresh_url,
            token_updater=self.token_saver
        )

    def request(self, url, req_type, json=None, data=None, params=None):
        """
        Base request using requests libary.

        `Args:`
            url: str
                The url request string; if ``url`` is a relative URL, it will be joined with
                the ``uri`` of the ``OAuthAPIConnector`; if ``url`` is an absolute URL, it will
                be used as is.
            req_type: str
                The request type. One of GET, POST, PATCH, DELETE, OPTIONS
            json: dict
                The payload of the request object. By using json, it will automatically
                serialize the dictionary
            data: str or byte or dict
                The payload of the request object. Use instead of json in some instances.
            params: dict
                The parameters to append to the url (e.g. http://myapi.com/things?id=1)

        `Returns:`
            requests response
        """
        full_url = urllib.parse.urljoin(self.uri, url)
        return self.client.request(
            req_type, full_url, headers=self.headers, auth=self.auth,
            json=json, data=data, params=params
        )

    def token_saver(self, token):
        self.token = token



from parsons.utilities.api_connector import APIConnector
from parsons.utilities import check_env
from parsons.bluelink.person import BluelinkPerson
import logging
import json

logger = logging.getLogger(__name__)

API_URL = 'https://api.bluelink.org/webhooks/'


class Bluelink:
    """
    Instantiate a Bluelink connector.
    Allows for a simple method of inserting person data to Bluelink via a webhook.
    # see: https://bluelinkdata.github.io/docs/BluelinkApiGuide#webhook

    `Args:`:
        user: str
            Bluelink webhook user name.
        password: str
            Bluelink webhook password.
    """
    def __init__(self, user=None, password=None):
        self.user = check_env.check('BLUELINK_WEBHOOK_USER', user)
        self.password = check_env.check('BLUELINK_WEBHOOK_PASSWORD', password)
        self.headers = {
            "Content-Type": "application/json",
        }
        self.api_url = API_URL
        self.api = APIConnector(self.api_url, auth=(self.user, self.password), headers=self.headers)

    def upsert_person(self, source, person=None):
        """
        Upsert a BluelinkPerson object into Bluelink.
        Rows will update, as opposed to being inserted, if an existing person record in
        Bluelink has a matching BluelinkIdentifier (same source and id) as the BluelinkPerson object
        passed into this function.

        `Args:`
            source: str
                String to identify that the data came from your system. For example,
                your company name.
            person: BluelinkPerson
                A BluelinkPerson object.
                Will be inserted to Bluelink, or updated if a matching record is found.
        `Returns:`
            int
            An http status code from the http post request to the Bluelink webhook.
        """
        data = {
            'source': source,
            'person': person
        }
        jdata = json.dumps(data,
                           default=lambda o: {k: v for k, v in o.__dict__.items() if v is not None})
        resp = self.api.post_request(url=self.api_url, data=jdata)
        return resp

    def bulk_upsert_person(self, source, tbl, row_to_person):
        """
        Upsert all rows into Bluelink, using the row_to_person function to
        transform rows to BluelinkPerson objects.

        `Args:`
            source: str
                String to identify that the data came from your system.
                For example, your company name.
            tbl: Table
                A parsons Table that represents people data.
            row_to_person: Callable[[dict],BluelinkPerson]
                A function that takes a dict representation of a row from the passed in tbl
                and returns a BluelinkPerson object.

        `Returns:`
            list[int]
            A list of https response status codes, one response for each row in the table.
        """
        people = BluelinkPerson.from_table(tbl, row_to_person)
        responses = []
        for person in people:
            response = self.upsert_person(source, person)
            responses.append(response)
        return responses



"""NGPVAN Bulk Import Endpoints"""
from parsons.etl.table import Table
from parsons.utilities import cloud_storage

import logging
import uuid
import csv

logger = logging.getLogger(__name__)


class BulkImport(object):

    def __init__(self):

        pass

    def get_bulk_import_resources(self):
        """
        Get bulk import resources that available to the user. These define
        the types of bulk imports that you can run. These might include
        ``Contacts``, ``ActivistCodes``, ``ContactsActivistCodes`` and others.

        `Returns:`
            list
                A list of resources.
        """

        r = self.connection.get_request('bulkImportJobs/resources')
        logger.info(f'Found {len(r)} bulk import resources.')
        return r

    def get_bulk_import_job(self, job_id):
        """
        Get a bulk import job status.

        `Args:`
            job_id : int
                The bulk import job id.
        `Returns:`
            dict
                The bulk import job
        """

        r = self.connection.get_request(f'bulkImportJobs/{job_id}')
        logger.info(f'Found bulk import job {job_id}.')
        return r

    def get_bulk_import_job_results(self, job_id):
        """
        Get result file of a bulk upload job. This will include one row
        per record processed as well as the status of each.

        If the job results have not been generated, either due to an error in the
        process or the fact the job is still processing, it will return ``None``.

        `Args:`
            job_id: int
                The bulk import job id.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        r = self.get_bulk_import_job(job_id)
        logger.info(f"Bulk Import Job Status: {r['status']}")
        if r['status'] == 'Completed':
            return Table.from_csv(r['resultFiles'][0]['url'])

        return None

    def get_bulk_import_mapping_types(self):
        """
        Get bulk import mapping types.

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        tbl = Table(self.connection.get_request('bulkImportMappingTypes'))
        logger.info(f'Found {tbl.num_rows} bulk import mapping types.')
        return tbl

    def get_bulk_import_mapping_type(self, type_name):
        """
        Get a single bulk import mapping type.

        `Args:`
            type_name: str
        `Returns`:
            dict
                A mapping type json
        """

        r = self.connection.get_request(f'bulkImportMappingTypes/{type_name}')
        logger.info(f'Found {type_name} bulk import mapping type.')
        return r

    def get_bulk_import_mapping_type_fields(self, type_name, field_name):
        """
        Get data about a field in a mapping type.

        `Args:`
            type_name: str
                The mapping type name
            field_name: str
                The field name
        `Returns:`
            dict
                A mapping type fields json
        """

        r = self.connection.get_request(f'bulkImportMappingTypes/{type_name}/{field_name}/values')
        logger.info(f'Found {type_name} bulk import mapping type field values.')
        return r

    def post_bulk_import(self, tbl, url_type, resource_type, mapping_types,
                         description, result_fields=None, **url_kwargs):
        # Internal method to post bulk imports.

        # Move to cloud storage
        file_name = str(uuid.uuid1())
        url = cloud_storage.post_file(tbl,
                                      url_type,
                                      file_path=file_name + '.zip',
                                      quoting=csv.QUOTE_ALL,
                                      **url_kwargs)
        logger.info(f'Table uploaded to {url_type}.')

        # Generate request json
        json = {"description": description,
                "file": {
                    "columnDelimiter": 'csv',
                    "columns": [{'name': c} for c in tbl.columns],
                    "fileName": file_name + '.csv',
                    "hasHeader": "True",
                    "hasQuotes": "True",
                    "sourceUrl": url},
                "actions": [{"resultFileSizeKbLimit": 5000,
                             "resourceType": resource_type,
                             "actionType": "loadMappedFile",
                             "mappingTypes": mapping_types}]
                }

        if result_fields:
            result_fields = [{'name': c} for c in result_fields]
            json['actions'][0]['columnsToIncludeInResultsFile'] = result_fields

        r = self.connection.post_request('bulkImportJobs', json=json)
        logger.info(f"Bulk upload {r['jobId']} created.")
        return r['jobId']

    def bulk_apply_activist_codes(self, tbl, url_type, **url_kwargs):
        """
        Bulk apply activist codes.

        The table may include the following columns. The first column
        must be ``vanid``.

        .. list-table::
            :widths: 25 25 50
            :header-rows: 1

            * - Column Name
              - Required
              - Description
            * - ``vanid``
              - Yes
              - A valid VANID primary key
            * - ``activistcodeid``
              - Yes
              - A valid activist code id
            * - ``datecanvassed``
              - No
              - An ISO formatted date
            * - ``contacttypeid``
              - No
              - The method of contact.

        `Args:`
            table: Parsons table
                A Parsons table.
            url_type: str
                The cloud file storage to use to post the file (``S3`` or ``GCS``).
                See :ref:`Cloud Storage <cloud-storage>` for more details.
            **url_kwargs: kwargs
                Arguments to configure your cloud storage url type. See
                :ref:`Cloud Storage <cloud-storage>` for more details.
        `Returns:`
            int
                The bulk import job id
        """

        return self.post_bulk_import(tbl,
                                     url_type,
                                     'ContactsActivistCodes',
                                     [{"name": "ActivistCode"}],
                                     'Activist Code Upload',
                                     **url_kwargs)

    def bulk_upsert_contacts(self, tbl, url_type, result_fields=None, **url_kwargs):
        """
        Bulk create or update contact records. Provide a Parsons table of contact data to
        create or update records.

        .. note::
            * The first column of the table must be VANID.
            * The other columns can be a combination of the columns listed below.
              The valid column names also accept permutations with underscores, spaces
              and capitalization (e.g. ``phonenumber`` = ``Phone_Number``).

        **Table Fields**

        .. list-table::
            :widths: 500 100 10
            :header-rows: 1

            * - Column
              - Valid Column Names
              - Notes
            * - VANID
              - ``vanid``
              -
            * - Voter VAN ID
              - ``votervanid``
              - The contact's MyVoters VANID
            * - External ID
              - ``externalid``, ``id``, ``pk``, ``voterbaseid``
              - An external id to be stored.
            * - **PII**
              -
              -
            * - First Name
              - ``fn``, ``firstname``, ``last``
              -
            * - Middle Name
              - ``mn``, ``middlename``, ``middle``
              -
            * - Last Name
              - ``ln``, ``lastname``, ``last``
              -
            * - Date of Birth
              - ``dob``, ``dateofbirth`` ``birthdate``
              - What type of thing does this need?
            * - Sex
              - ``sex``, ``gender``
              -
            * - **Physical Address**
              -
              -
            * - Address Line 1
              - ``addressline1``, ``address1``, ``address``
              -
            * - Address Line 2
              - ``addressline2``, ``address2``
              -
            * - Address Line 3
              - ``addressline3``, ``address3``
              -
            * - City
              - ``city``
              -
            * - State Or Province
              - ``state``, ``st``, ``stateorprovince``
              -
            * - Country Code
              - ``countrycode``, ``country``
              - A valid two character country code (e.g. ``US``)
            * - Display As Entered
              - ``displayasentered``
              - Required values are ``Y`` and ``N``. Determines if the address is
                processed through address correction.
            * - **Phones**
              -
              -
            * - Cell Phone
              - ``cellphone``, ``cell``
              -
            * - Cell Phone Country Code
              - ``cellcountrycode``, ``cellphonecountrycode``
              - A valid two digit country code (e.g. ``01``)
            * - Home Phone
              - ``homephone``, ``home``, ``phone``
              -
            * - Home Phone Country Code
              - ``homecountrycode``, ``homephonecountrycode``
              -
            * - **Email**
              -
              -
            * - Email
              - ``email``, ``emailaddress``
              -

        `Args:`
            table: Parsons table
              A Parsons table.
            url_type: str
              The cloud file storage to use to post the file. Currently only ``S3``.
            results_fields: list
              A list of fields to include in the results file.
            **url_kwargs: kwargs
                Arguments to configure your cloud storage url type. See
                :ref:`Cloud Storage <cloud-storage>` for more details.
        `Returns:`
            int
                The bulk import job id
        """

        tbl = tbl.map_columns(COLUMN_MAP, exact_match=False)

        return self.post_bulk_import(tbl,
                                     url_type,
                                     'Contacts',
                                     [{'name': 'CreateOrUpdateContact'}],
                                     'Create Or Update Contact Records',
                                     result_fields=result_fields,
                                     **url_kwargs)


# This is a column mapper that is used to accept additional column names and provide
# flexibility for the user.

COLUMN_MAP = {'firstname': ['fn', 'first'],
              'middlename': ['mn', 'middle'],
              'lastname': ['ln', 'last'],
              'dob': ['dateofbirth', 'birthdate'],
              'sex': ['gender'],
              'addressline1': ['address', 'addressline1', 'address1'],
              'addressline2': ['addressline2', 'address2'],
              'addressline3': ['addressline3', 'address3'],
              'city': [],
              'stateorprovince': ['state', 'st'],
              'countrycode': ['country'],
              'displayasentered': [],
              'cellphone': ['cell'],
              'cellphonecountrycode': ['cellcountrycode'],
              'phone': ['home', 'homephone'],
              'phonecountrycode': ['phonecountrycode'],
              'email': ['emailaddress']}



from contextlib import contextmanager
from stat import S_ISDIR, S_ISREG
import logging
import paramiko
import re

from parsons.utilities import files as file_utilities
from parsons.etl import Table
from parsons.sftp.utilities import connect

logger = logging.getLogger(__name__)


class SFTP(object):
    """
    Instantiate SFTP Class

    `Args:`
        host: str
            The host name
        username: str
            The user name
        password: str
            The password
        rsa_private_key_file str
            Absolute path to a private RSA key used
            to authenticate stfp connection
        port: int
            Specify if different than the standard port 22
    `Returns:`
        SFTP Class
    """

    def __init__(self, host, username, password, port=22, rsa_private_key_file=None):
        self.host = host
        if not self.host:
            raise ValueError("Missing the SFTP host name")

        self.username = username
        if not self.username:
            raise ValueError("Missing the SFTP username")

        if not (password or rsa_private_key_file):
            raise ValueError("Missing password or SSH authentication key")

        self.password = password
        self.rsa_private_key_file = rsa_private_key_file
        self.port = port

    @contextmanager
    def create_connection(self):
        """
        Create an SFTP connection.

        Returns:
            SFTP Connection object
        """

        transport = paramiko.Transport((self.host, self.port))
        pkey = None
        if self.rsa_private_key_file:
            # we need to read it in
            pkey = paramiko.RSAKey.from_private_key_file(self.rsa_private_key_file)

        transport.connect(username=self.username, password=self.password, pkey=pkey)
        conn = paramiko.SFTPClient.from_transport(transport)
        yield conn
        conn.close()
        transport.close()

    def list_directory(self, remote_path='.', connection=None):
        """
        List the contents of a directory

        `Args:`
            remote_path: str
                The remote path of the directory
            connection: obj
                An SFTP connection object
        `Returns:`
            list of files and subdirectories in the provided directory
        """

        if connection:
            return connection.listdir(path=remote_path)
        else:
            with self.create_connection() as connection:
                return connection.listdir(path=remote_path)

    def make_directory(self, remote_path, connection=None):
        """
        Makes a new directory on the SFTP server

        `Args:`
            remote_path: str
                The remote path of the directory
            connection: obj
                An SFTP connection object
        """

        if connection:
            connection.mkdir(remote_path)
        else:
            with self.create_connection() as connection:
                connection.mkdir(remote_path)

    def remove_directory(self, remote_path, connection=None):
        """
        Remove a directory from the SFTP server

        `Args:`
            remote_path: str
                The remote path of the directory
            connection: obj
                An SFTP connection object
        """

        if connection:
            connection.rmdir(remote_path)
        else:
            with self.create_connection() as connection:
                connection.rmdir(remote_path)

    def get_file(self, remote_path, local_path=None, connection=None):
        """
        Download a file from the SFTP server

        `Args:`
            remote_path: str
                The remote path of the file to download
            local_path: str
                The local path where the file will be downloaded. If not specified, a temporary
                file will be created and returned, and that file will be removed automatically
                when the script is done running.
            connection: obj
                An SFTP connection object
        `Returns:`
            str
                The path of the local file
        """

        if not local_path:
            local_path = file_utilities.create_temp_file_for_path(remote_path)

        if connection:
            connection.get(remote_path, local_path)
        else:
            with self.create_connection() as connection:
                connection.get(remote_path, local_path)

        return local_path

    @connect
    def get_files(self, files_to_download=None, remote=None, connection=None, pattern=None,
                  local_paths=None):
        """
        Download a list of files, either by providing the list explicitly, providing directories
        that contain files to download, or both.

        `Args:`
            files_to_download: list
                A list of full remote paths (can be relative) to files to download
            remote: str or list
                A path to a remote directory or a list of paths
            connection: obj
                An SFTP connection object
            pattern: str
                A regex pattern with which to select file names. Defaults to None, in which case
                all files will be selected.
            local_paths: list
                A list of paths to which to save the selected files. Defaults to None. If it is not
                the same length as the files to be fetched, temporary files are used instead.
        `Returns:`
            list
                Local paths where the files are saved.
        """

        if not (files_to_download or remote):
            raise ValueError("You must provide either `files_to_download`, `remote`, or both, as "
                             "an argument to `get_files`.")

        if not files_to_download:
            files_to_download = []

        if remote:
            try:  # assume `remote` is a str
                files_to_download.extend(self.list_files(remote, connection, pattern))
            except TypeError:  # if it's not a str it's a list
                files_to_download.extend(
                    f for file_list in [self.list_files(directory, connection, pattern)
                                        for directory in remote]
                    for f in file_list
                )

        if local_paths and len(local_paths) != len(files_to_download):
            logger.warning("You provided a list of local paths for your files but it was not "
                           "the same length as the files you are going to download.\nDefaulting to "
                           "temporary files.")
            local_paths = []

        if local_paths:
            return [
                self.get_file(remote_path, local_path, connection)
                for local_path, remote_path in zip(local_paths, files_to_download)
            ]

        else:
            return [
                self.get_file(file, local_path=None, connection=connection)
                for file in files_to_download
            ]

    def get_table(self, remote_path, connection=None):
        """
        Download a csv from the server and convert into a Parsons table.

        The file may be compressed with gzip, or zip, but may not contain
        multiple files in the archive.

        `Args:`
            remote_path: str
                The remote path of the file to download
            connection: obj
                An SFTP connection object
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        if not file_utilities.valid_table_suffix(remote_path):
            raise ValueError('File type cannot be converted to a Parsons table.')

        return Table.from_csv(self.get_file(remote_path, connection=connection))

    def put_file(self, local_path, remote_path, connection=None):
        """
        Put a file on the SFTP server

        `Args:`
            local_path: str
                The local path of the source file
            remote_path: str
                The remote path of the new file
            connection: obj
                An SFTP connection object
        """
        if connection:
            connection.put(local_path, remote_path)
        with self.create_connection() as connection:
            connection.put(local_path, remote_path)

    def remove_file(self, remote_path, connection=None):
        """
        Delete a file on the SFTP server

        `Args:`
            remote_path: str
                The remote path of the file
            connection: obj
                An SFTP connection object
        """

        if connection:
            connection.remove(remote_path)
        with self.create_connection() as connection:
            connection.remove(remote_path)

    def get_file_size(self, remote_path, connection=None):
        """
        Get the size of a file in MB on the SFTP server. The file is
        not downloaded locally.

        `Args:`
            remote_path: str
                The remote path of the file
            connection: obj
                An SFTP connection object
        `Returns:`
            int
                The file size in MB.
        """

        if connection:
            size = connection.file(remote_path, 'r')._get_size()
        else:
            with self.create_connection() as connection:
                size = connection.file(remote_path, 'r')._get_size()

        return size / 1024

    @staticmethod
    def _list_contents(remote_path, connection, dir_pattern=None, file_pattern=None):

        dirs_to_return = []
        files_to_return = []
        dirs_and_files = [(S_ISDIR, dir_pattern, True, dirs_to_return),
                          (S_ISREG, file_pattern, False, files_to_return)]

        try:
            for entry in connection.listdir_attr(remote_path):
                entry_pathname = remote_path + "/" + entry.filename
                for method, pattern, do_search_full_path, paths in dirs_and_files:
                    string = entry_pathname if do_search_full_path else entry.filename
                    if method(entry.st_mode) and (not pattern or re.search(pattern, string)):
                        paths.append(entry_pathname)
        except FileNotFoundError:  # This error is raised when a directory is empty
            pass
        return dirs_to_return, files_to_return

    @connect
    def list_subdirectories(self, remote_path, connection=None, pattern=None):
        """
        List the subdirectories of a directory on the remote server.

        `Args:`
            remote_path: str
                The remote directory whose subdirectories will be listed
            connection: obj
                An SFTP connection object
            pattern: str
                A regex pattern with which to select full directory paths. Defaults to None, in
                which case all subdirectories will be selected.
        `Returns:`
            list
                The subdirectories in `remote_path`.
        """
        return self._list_contents(remote_path, connection, dir_pattern=pattern)[0]

    @connect
    def list_files(self, remote_path, connection=None, pattern=None):
        """
        List the files in a directory on the remote server.

        `Args:`
            remote_path: str
                The remote directory whose files will be listed
            connection: obj
                An SFTP connection object
            pattern: str
                A regex pattern with which to select file names. Defaults to None, in which case
                all files will be selected.
        `Returns:`
            list
                The files in `remote_path`.
        """
        return self._list_contents(remote_path, connection, file_pattern=pattern)[1]

    @connect
    def walk_tree(self, remote_path, connection=None, download=False, dir_pattern=None,
                  file_pattern=None, max_depth=2):
        """
        Recursively walks a directory, fetching all subdirectories and files (as long as
        they match `dir_pattern` and `file_pattern`, respectively) and the maximum directory
        depth hasn't been exceeded. Optionally downloads discovered files.

        `Args:`
            remote_path: str
                The top level directory to walk
            connection: obj
                An SFTP connection object
            download: bool
                Whether to download discovered files
            dir_pattern: str
                A regex pattern with which to select directories. Defaults to None, in which case
                all directories will be selected.
            file_pattern: str
                A regex pattern with which to select files. Defaults to None, in which case
                all files will be selected.
            max_depth: int
                A limit on how many directories deep to traverse.  The default, 2, will search the
                contents of `remote_path` and its subdirectories.
        `Returns:`
            tuple
                A list of directories touched and a list of files.  If the files were downloaded
                the file list will consist of local paths, if not, remote paths.
        """

        if max_depth > 3:
            logger.warning("Calling `walk_tree` with `max_depth` {}.  "
                           "Recursively walking a remote directory will be much slower than a "
                           "similar operation on a local file system.".format(max_depth))

        to_return = self._walk_tree(remote_path, connection, download, dir_pattern, file_pattern,
                                    max_depth=max_depth)

        return to_return

    def _walk_tree(self, remote_path, connection, download=False, dir_pattern=None,
                   file_pattern=None, depth=0, max_depth=2):

        dir_list = []
        file_list = []

        depth += 1

        dirs, files = self._list_contents(remote_path, connection, dir_pattern, file_pattern)

        if download:
            self.get_files(files_to_download=files)

        if depth < max_depth:
            for directory in dirs:
                deeper_dirs, deeper_files = self._walk_tree(
                    directory, connection, download, dir_pattern, file_pattern, depth, max_depth
                )
                dir_list.extend(deeper_dirs)
                file_list.extend(deeper_files)

        dir_list.extend(dirs)
        file_list.extend(files)

        return dir_list, file_list



import os
import unittest
from unittest.mock import patch

import requests_mock
from parsons import Table, GitHub
from github.GithubException import UnknownObjectException
from parsons.github.github import ParsonsGitHubError


_dir = os.path.dirname(__file__)


class TestGitHub(unittest.TestCase):

    def setUp(self):
        self.github = GitHub(access_token='token')

    @requests_mock.Mocker()
    def test_wrap_github_404(self, m):
        with patch("github.Github.get_repo") as get_repo_mock:
            get_repo_mock.side_effect = UnknownObjectException("", "")
            with self.assertRaises(ParsonsGitHubError):
                self.github.get_repo("octocat/Hello-World")

    @requests_mock.Mocker()
    def test_get_repo(self, m):
        with open(os.path.join(_dir, 'test_data', 'test_get_repo.json'), 'r') as f:
            m.get(requests_mock.ANY, text=f.read())
        repo = self.github.get_repo('octocat/Hello-World')
        self.assertEqual(repo['id'], 1296269)
        self.assertEqual(repo['name'], 'Hello-World')

    @requests_mock.Mocker()
    def test_list_repo_issues(self, m):
        with open(os.path.join(_dir, 'test_data', 'test_get_repo.json'), 'r') as f:
            m.get('https://api.github.com:443/repos/octocat/Hello-World', text=f.read())
        with open(os.path.join(_dir, 'test_data', 'test_list_repo_issues.json'), 'r') as f:
            m.get('https://api.github.com:443/repos/octocat/Hello-World/issues', text=f.read())
        issues_table = self.github.list_repo_issues('octocat/Hello-World')
        self.assertIsInstance(issues_table, Table)
        self.assertEqual(len(issues_table.table), 2)
        self.assertEqual(issues_table[0]['id'], 1)
        self.assertEqual(issues_table[0]['title'], 'Found a bug')

    @requests_mock.Mocker()
    def test_download_file(self, m):
        with open(os.path.join(_dir, 'test_data', 'test_get_repo.json'), 'r') as f:
            m.get('https://api.github.com:443/repos/octocat/Hello-World', text=f.read())
        with open(os.path.join(_dir, 'test_data', 'test_download_file.csv'), 'r') as f:
            m.get(
                'https://raw.githubusercontent.com/octocat/Hello-World/testing/data.csv',
                text=f.read()
            )

        file_path = self.github.download_file('octocat/Hello-World', 'data.csv', branch='testing')
        with open(file_path, 'r') as f:
            file_contents = f.read()

        self.assertEqual(file_contents, "header\ndata\n")



from parsons import Auth0, Table
from test.utils import assert_matching_tables
import requests_mock
import unittest

CLIENT_ID = 'abc'
CLIENT_SECRET = 'def'
DOMAIN = 'fakedomain.auth0.com'


class TestAuth0(unittest.TestCase):
    def setUp(self):
        self.auth0 = Auth0(CLIENT_ID, CLIENT_SECRET, DOMAIN)

    @requests_mock.Mocker()
    def test_delete_user(self, m):
        user_id = 1
        m.delete(f'{self.auth0.base_url}/api/v2/users/{user_id}', status_code=204)
        self.assertEqual(self.auth0.delete_user(user_id), 204)

    @requests_mock.Mocker()
    def test_get_users_by_email(self, m):
        email = 'fakeemail@fakedomain.com'
        mock_users = [{'email': 'fake3mail@fakedomain.com', 'id': 2}]
        m.get(f'{self.auth0.base_url}/api/v2/users-by-email?email={email}', json=mock_users)
        assert_matching_tables(self.auth0.get_users_by_email(email), Table(mock_users), True)



import unittest
import os
import requests_mock
import unittest.mock as mock
from parsons import VAN, Table
from test.utils import assert_matching_tables
from parsons.utilities import cloud_storage

os.environ['VAN_API_KEY'] = 'SOME_KEY'


class TestBulkImport(unittest.TestCase):

    def setUp(self):

        self.van = VAN(os.environ['VAN_API_KEY'],
                       db="MyVoters", raise_for_status=False)

    @requests_mock.Mocker()
    def test_get_bulk_import_resources(self, m):

        json = ['Contacts', 'Contributions',
                'ActivistCodes', 'ContactsActivistCodes']

        m.get(self.van.connection.uri + 'bulkImportJobs/resources', json=json)

        self.assertEqual(self.van.get_bulk_import_resources(), json)

    @requests_mock.Mocker()
    def test_get_bulk_import_job(self, m):

        m.get(self.van.connection.uri +
              'bulkImportJobs/53407', json=bulk_import_job)

        self.assertEqual(self.van.get_bulk_import_job(53407), bulk_import_job)

    @requests_mock.Mocker()
    def test_get_bulk_import_job_results(self, m):

        results_tbl = Table([['BulkUploadDataID', 'ULFileID', 'PrimaryKey',
                              'PrimaryKeyType', 'MailingAddress_3581'],
                             ['1', '1983', '101596008', 'VanID', 'Processed']])

        bulk_import_job = {'id': 92,
                           'status': 'Completed',
                           'resourceType': 'Contacts',
                           'webhookUrl': None,
                           'resultFileSizeLimitKb': 5000,
                           'errors': [],
                           'resultFiles': [{
                               'url': Table.to_csv(results_tbl),
                               'dateExpired': '2020-09-04T22:07:04.0770295-04:00'
                           }]
                           }

        m.get(self.van.connection.uri +
              'bulkImportJobs/53407', json=bulk_import_job)
        assert_matching_tables(
            self.van.get_bulk_import_job_results(53407), results_tbl)

    @requests_mock.Mocker()
    def test_get_bulk_import_mapping_types(self, m):

        m.get(self.van.connection.uri +
              'bulkImportMappingTypes', json=mapping_type)

        assert_matching_tables(
            self.van.get_bulk_import_mapping_types(), Table(mapping_type))

    @requests_mock.Mocker()
    def test_get_bulk_import_mapping_type(self, m):

        m.get(self.van.connection.uri +
              'bulkImportMappingTypes/ActivistCode', json=mapping_type)

        self.assertEqual(self.van.get_bulk_import_mapping_type(
            'ActivistCode'), mapping_type)

    @requests_mock.Mocker()
    def get_bulk_import_mapping_type_fields(self, m):

        json = [{'name': 'Unsubscribed', 'id': '0', 'parents': None},
                {'name': 'Not Subscribed', 'id': '1', 'parents': None},
                {'name': 'Subscribed', 'id': '2', 'parents': None}]
        m.get(self.van.connection.uri +
              'bulkImportMappingTypes/Email/EmailSubscriptionStatusId/values')

        r = self.van.get_bulk_import_mapping_type_fields(
            'Email', 'EmailSubscriptionStatusId')
        self.assertEqual(json, r)

    @requests_mock.Mocker()
    def test_post_bulk_import(self, m):

        # Mock Cloud Storage
        cloud_storage.post_file = mock.MagicMock()
        cloud_storage.post_file.return_value = 'https://s3.com/my_file.zip'

        tbl = Table([['Vanid', 'ActivistCodeID'], [1234, 345345]])

        m.post(self.van.connection.uri +
               'bulkImportJobs', json={'jobId': 54679})

        r = self.van.post_bulk_import(tbl,
                                      'S3',
                                      'ContactsActivistCodes',
                                      [{"name": "ActivistCode"}],
                                      'Activist Code Upload',
                                      bucket='my-bucket')

        self.assertEqual(r, 54679)

    @requests_mock.Mocker()
    def test_bulk_apply_activist_codes(self, m):

        # Mock Cloud Storage
        cloud_storage.post_file = mock.MagicMock()
        cloud_storage.post_file.return_value = 'https://s3.com/my_file.zip'

        tbl = Table([['Vanid', 'ActivistCodeID'], [1234, 345345]])

        m.post(self.van.connection.uri +
               'bulkImportJobs', json={'jobId': 54679})

        job_id = self.van.bulk_apply_activist_codes(
            tbl, url_type="S3", bucket='my-bucket')

        self.assertEqual(job_id, 54679)

    @requests_mock.Mocker()
    def test_bulk_upsert_contacts(self, m):

        # Mock Cloud Storage
        cloud_storage.post_file = mock.MagicMock()
        cloud_storage.post_file.return_value = 'https://s3.com/my_file.zip'

        tbl = Table([['Vanid', 'email'], [1234, 'me@me.com']])

        m.post(self.van.connection.uri +
               'bulkImportJobs', json={'jobId': 54679})

        job_id = self.van.bulk_upsert_contacts(
            tbl, url_type="S3", bucket='my-bucket')

        self.assertEqual(job_id, 54679)


mapping_type = {'name': 'ActivistCode',
                'displayName': 'Apply Activist Code',
                'allowMultipleMode': 'Multiple',
                'resourceTypes': ['Contacts', 'ContactsActivistCodes'],
                'fields': [{
                    'name': 'ActivistCodeID',
                    'description': 'Activist Code ID',
                    'hasPredefinedValues': True,
                    'isRequired': True,
                    'canBeMappedToColumn': True,
                    'canBeMappedByName': True,
                    'parents': None},
                    {'name': 'CanvassedBy',
                     'description': 'Recruited By, Must be a Valid User ID',
                     'hasPredefinedValues': False,
                     'isRequired': False,
                     'canBeMappedToColumn': True,
                     'canBeMappedByName': True,
                     'parents': None},
                    {'name': 'DateCanvassed',
                     'description': 'Contacted When',
                     'hasPredefinedValues': False,
                     'isRequired': False,
                     'canBeMappedToColumn': True,
                     'canBeMappedByName': True,
                     'parents': [{
                         'parentFieldName': 'CanvassedBy',
                         'limitedToParentValues': None
                     }]
                     }, {
                    'name': 'ContactTypeID',
                    'description': 'Contacted How',
                    'hasPredefinedValues': True,
                    'isRequired': False,
                    'canBeMappedToColumn': True,
                    'canBeMappedByName': True,
                    'parents': [{
                        'parentFieldName': 'CanvassedBy',
                        'limitedToParentValues': None
                    }]
                }]

                }

bulk_import_job = {'id': 92,
                   'status': 'Completed',
                   'resourceType': 'Contacts',
                   'webhookUrl': None,
                   'resultFileSizeLimitKb': 5000,
                   'errors': [],
                   'resultFiles': [{
                       'url': 'https://ngpvan.com/bulk-import-jobs/f023.csv',
                       'dateExpired': '2020-09-04T22:07:04.0770295-04:00'
                   }]
                   }



from parsons import Table, Shopify
from test.utils import assert_matching_tables
import requests_mock
import unittest

SUBDOMAIN = 'myorg'
PASSWORD = 'abc123'
API_KEY = 'abc123'
API_VERSION = '2020-10'


class TestShopify(unittest.TestCase):
    mock_count_all = {
        'count': 2
    }
    mock_count_date = mock_count_since = {
        'count': 1
    }
    mock_graphql = {
        'data': {
            'orders': {
                'edges': [{
                    'node': {
                        'id': 1
                    }
                }]
            }
        }
    }
    mock_orders_all = {
        'orders': [{
            'created_at': '2020-10-19T12:00:00-04:00',
            'financial_status': 'paid',
            'id': 1
        }, {
            'created_at': '2020-10-20T12:00:00-04:00',
            'financial_status': 'refunded',
            'id': 2
        }]
    }
    mock_orders_completed = {
        'orders': [{
            'created_at': '2020-10-19T12:00:00-04:00',
            'financial_status': 'paid',
            'id': 1
        }]
    }
    mock_orders_date = mock_orders_since = {
        'orders': [{
            'created_at': '2020-10-20T12:00:00-04:00',
            'financial_status': 'refunded',
            'id': 2
        }]
    }
    mock_result_all = Table([('created_at', 'financial_status', 'id'),
                             ('2020-10-19T12:00:00-04:00', 'paid', 1), ('2020-10-20T12:00:00-04:00',
                                                                        'refunded', 2)])
    mock_result_completed = Table([('created_at', 'financial_status', 'id'),
                                   ('2020-10-19T12:00:00-04:00', 'paid', 1)])
    mock_result_date = mock_result_since = Table([('created_at', 'financial_status', 'id'),
                                                  ('2020-10-20T12:00:00-04:00', 'refunded', 2)])

    def setUp(self):
        self.shopify = Shopify(SUBDOMAIN, PASSWORD, API_KEY, API_VERSION)

    @requests_mock.Mocker()
    def test_get_count(self, m):
        m.get(self.shopify.get_query_url(None, None, "orders", True), json=self.mock_count_all)
        m.get(self.shopify.get_query_url('2020-10-20', None, "orders", True),
              json=self.mock_count_date)
        m.get(self.shopify.get_query_url(None, 2, "orders", True), json=self.mock_count_since)
        self.assertEqual(self.shopify.get_count(None, None, "orders"), 2)
        self.assertEqual(self.shopify.get_count('2020-10-20', None, "orders"), 1)
        self.assertEqual(self.shopify.get_count(None, 2, "orders"), 1)

    @requests_mock.Mocker()
    def test_get_orders(self, m):
        m.get(self.shopify.get_query_url(None, None, 'orders', False), json=self.mock_orders_all)
        m.get(self.shopify.get_query_url('2020-10-20', None, 'orders', False),
              json=self.mock_orders_date)
        m.get(self.shopify.get_query_url(None, 2, 'orders', False), json=self.mock_orders_since)
        m.get(self.shopify.get_query_url(None, None, 'orders', False) + '&financial_status=paid',
              json=self.mock_orders_completed)
        assert_matching_tables(self.shopify.get_orders(None, None, False), self.mock_result_all)
        assert_matching_tables(self.shopify.get_orders('2020-10-20', None, False),
                               self.mock_result_date)
        assert_matching_tables(self.shopify.get_orders(None, 2, False), self.mock_result_since)
        assert_matching_tables(self.shopify.get_orders(None, None, True),
                               self.mock_result_completed)

    @requests_mock.Mocker()
    def test_get_query_url(self, m):
        self.assertEqual(self.shopify.get_query_url(None, None, "orders", True),
                         f'https://{SUBDOMAIN}.myshopify.com/admin/api/{API_VERSION}/orders/' +
                         'count.json?limit=250&status=any')
        self.assertEqual(self.shopify.get_query_url('2020-10-20', None, "orders", True),
                         f'https://{SUBDOMAIN}.myshopify.com/admin/api/{API_VERSION}/orders/' +
                         'count.json?limit=250&status=any&created_at_min=2020-10-20T00:00:00&' +
                         'created_at_max=2020-10-21T00:00:00')
        self.assertEqual(self.shopify.get_query_url(None, 2, "orders", True),
                         f'https://{SUBDOMAIN}.myshopify.com/admin/api/{API_VERSION}/orders/' +
                         'count.json?limit=250&status=any&since_id=2')
        self.assertEqual(self.shopify.get_query_url(None, None, "orders", False),
                         f'https://{SUBDOMAIN}.myshopify.com/admin/api/{API_VERSION}/orders.json?' +
                         'limit=250&status=any')

    @requests_mock.Mocker()
    def test_graphql(self, m):
        m.post(
            'https://{0}.myshopify.com/admin/api/{1}/graphql.json'.format(SUBDOMAIN, API_VERSION),
            json=self.mock_graphql
        )
        self.assertEqual(self.shopify.graphql("""
            {{
                orders(query: "financial_status:=paid", first: 100) {{
                    edges {{
                        node {{
                            id
                        }}
                    }}
                }}
            }}
        """), self.mock_graphql['data'])



# These are reserved words by Redshift and cannot be used as column names.
RESERVED_WORDS = [
    "AES128", "AES256", "ALL", "ALLOWOVERWRITE", "ANALYSE", "ANALYZE", "AND",
    "ANY", "ARRAY", "AS", "ASC", "AUTHORIZATION", "BACKUP", "BETWEEN", "BINARY",
    "BLANKSASNULL", "BOTH", "BYTEDICT", "BZIP2", "CASE", "CAST", "CHECK",
    "COLLATE", "COLUMN", "CONSTRAINT", "CREATE", "CREDENTIALS", "CROSS",
    "CURRENT_DATE", "CURRENT_TIME", "CURRENT_TIMESTAMP", "CURRENT_USER",
    "CURRENT_USER_ID", "DEFAULT", "DEFERRABLE", "DEFLATE", "DEFRAG", "DELTA",
    "DELTA32K", "DESC", "DISABLE", "DISTINCT", "DO", "ELSE", "EMPTYASNULL",
    "ENABLE", "ENCODE", "ENCRYPT", "ENCRYPTION", "END", "EXCEPT", "EXPLICIT",
    "FALSE", "FOR", "FOREIGN", "FREEZE", "FROM", "FULL", "GLOBALDICT256",
    "GLOBALDICT64K", "GRANT", "GROUP", "GZIP", "HAVING", "IDENTITY", "IGNORE",
    "ILIKE", "IN", "INITIALLY", "INNER", "INTERSECT", "INTO", "IS", "ISNULL",
    "JOIN", "LEADING", "LEFT", "LIKE", "LIMIT", "LOCALTIME", "LOCALTIMESTAMP",
    "LUN", "LUNS", "LZO", "LZOP", "MINUS", "MOSTLY13", "MOSTLY32", "MOSTLY8",
    "NATURAL", "NEW", "NOT", "NOTNULL", "NULL", "NULLS", "OFF", "OFFLINE",
    "OFFSET", "OLD", "ON", "ONLY", "OPEN", "OR", "ORDER", "OUTER", "OVERLAPS",
    "PARALLEL", "PARTITION", "PERCENT", "PERMISSIONS", "PLACING", "PRIMARY",
    "RAW", "READRATIO", "RECOVER", "REFERENCES", "RESPECT", "REJECTLOG",
    "RESORT", "RESTORE", "RIGHT", "SELECT", "SESSION_USER", "SIMILAR", "SOME",
    "SYSDATE", "SYSTEM", "TABLE", "TAG", "TDES", "TEXT255", "TEXT32K", "THEN",
    "TIMESTAMP", "TO", "TOP", "TRAILING", "TRUE", "TRUNCATECOLUMNS", "UNION",
    "UNIQUE", "USER", "USING", "VERBOSE", "WALLET", "WHEN", "WHERE", "WITH",
    "WITHOUT",
]

VARCHAR = "varchar"
FLOAT = "float"

DO_PARSE_BOOLS = False
BOOL = "bool"
TRUE_VALS = ("TRUE", "T", "YES", "Y", "1", 1)
FALSE_VALS = ("FALSE", "F", "NO", "N", "0", 0)

# The following values are the minimum and maximum values for MySQL int
# types. https://dev.mysql.com/doc/refman/8.0/en/integer-types.html
SMALLINT = "smallint"
SMALLINT_MIN = -32768
SMALLINT_MAX = 32767

MEDIUMINT = "mediumint"
MEDIUMINT_MIN = -8388608
MEDIUMINT_MAX = 8388607

INT = "int"
INT_MIN = -2147483648
INT_MAX = 2147483647

BIGINT = "bigint"

INT_TYPES = [SMALLINT, MEDIUMINT, INT, BIGINT]

COL_NAME_MAX_LEN = 100

IS_CASE_SENSITIVE = False

REPLACE_CHARS = {" ": "_"}



from parsons.etl.table import Table
from parsons.utilities import check_env
import requests


class Auth0(object):
    """
    Instantiate the Auth0 class

    `Args:`
        client_id: str
            The Auth0 client ID. Not required if ``AUTH0_CLIENT_ID`` env variable set.
        client_secret: str
            The Auth0 client secret. Not required if ``AUTH0_CLIENT_SECRET`` env variable set.
        domain: str
            The Auth0 domain. Not required if ``AUTH0_DOMAIN`` env variable set.
    `Returns:`
        Auth0 Class
    """
    def __init__(self, client_id=None, client_secret=None, domain=None):
        self.base_url = f"https://{check_env.check('AUTH0_DOMAIN', domain)}"
        access_token = requests.post(f'{self.base_url}/oauth/token', data={
            'grant_type': 'client_credentials',  # OAuth 2.0 flow to use
            'client_id': check_env.check('AUTH0_CLIENT_ID', client_id),
            'client_secret': check_env.check('AUTH0_CLIENT_SECRET', client_secret),
            'audience': f'{self.base_url}/api/v2/'
        }).json().get('access_token')
        self.headers = {
            'Authorization': f'Bearer {access_token}',
            'Content-Type': 'application/json'
        }

    def delete_user(self, id):
        """
        Delete Auth0 user.

        `Args:`
            id: str
                The user ID of the record to delete.
        `Returns:`
            int
        """
        return requests.delete(
            f'{self.base_url}/api/v2/users/{id}', headers=self.headers
        ).status_code

    def get_users_by_email(self, email):
        """
        Get Auth0 users by email.

        `Args:`
            email: str
                The user email of the record to get.
        `Returns:`
            Table Class
        """
        return Table(requests.get(
            f'{self.base_url}/api/v2/users-by-email', headers=self.headers, params={'email': email}
        ).json())



from parsons.bluelink.bluelink import Bluelink
from parsons.bluelink.person import BluelinkPerson, \
    BluelinkEmail, BluelinkAddress, BluelinkPhone, BluelinkIdentifier, BluelinkTag, BluelinkScore

__all__ = [
    'Bluelink',
    'BluelinkPerson', 'BluelinkEmail', 'BluelinkAddress', 'BluelinkPhone',
    'BluelinkIdentifier', 'BluelinkTag', 'BluelinkScore'
]



from requests.auth import HTTPBasicAuth
from parsons.etl import Table
from parsons.utilities import check_env
from parsons.utilities.api_connector import APIConnector
from parsons.utilities.datetime import date_to_timestamp
import logging

logger = logging.getLogger(__name__)

CAPITOL_CANARY_URI = 'https://api.phone2action.com/2.0/'


class CapitolCanary(object):
    """
    Instantiate CapitolCanary Class

    `Args:`
        app_id: str
            The CapitolCanary provided application id. Not required if ``CAPITOLCANARY_APP_ID``
            env variable set.
        app_key: str
            The CapitolCanary provided application key. Not required if ``CAPITOLCANARY_APP_KEY``
            env variable set.
    `Returns:`
        CapitolCanary Class
    """

    def __init__(self, app_id=None, app_key=None):
        # check first for CapitolCanary branded app key and ID
        cc_app_id = check_env.check('CAPITOLCANARY_APP_ID', None, optional=True)
        cc_app_key = check_env.check('CAPITOLCANARY_APP_KEY', None, optional=True)

        self.app_id = cc_app_id or check_env.check('PHONE2ACTION_APP_ID', app_id)
        self.app_key = cc_app_key or check_env.check('PHONE2ACTION_APP_KEY', app_key)
        self.auth = HTTPBasicAuth(self.app_id, self.app_key)
        self.client = APIConnector(CAPITOL_CANARY_URI, auth=self.auth)

    def _paginate_request(self, url, args=None, page=None):
        # Internal pagination method

        if page is not None:
            args['page'] = page

        r = self.client.get_request(url, params=args)

        json = r['data']

        if page is not None:
            return json

        # If count of items is less than the total allowed per page, paginate
        while r['pagination']['count'] == r['pagination']['per_page']:

            r = self.client.get_request(r['pagination']['next_url'], args)
            json.extend(r['data'])

        return json

    def get_advocates(self, state=None, campaign_id=None, updated_since=None, page=None):
        """
        Return advocates (person records).

        If no page is specified, the method will automatically paginate through the available
        advocates.

        `Args:`
            state: str
                Filter by US postal abbreviation for a state
                or territory e.g., "CA" "NY" or "DC"
            campaign_id: int
                Filter to specific campaign
            updated_since: str or int or datetime
                Fetch all advocates updated since the date provided; this can be a datetime
                object, a UNIX timestamp, or a date string (ex. '2014-01-05 23:59:43')
            page: int
                Page number of data to fetch; if this is specified, call will only return one
                page.
        `Returns:`
            A dict of parsons tables:
                * emails
                * phones
                * memberships
                * tags
                * ids
                * fields
                * advocates
        """

        # Convert the passed in updated_since into a Unix timestamp (which is what the API wants)
        updated_since = date_to_timestamp(updated_since)

        args = {'state': state,
                'campaignid': campaign_id,
                'updatedSince': updated_since}

        logger.info('Retrieving advocates...')
        json = self._paginate_request('advocates', args=args, page=page)

        return self._advocates_tables(Table(json))

    def _advocates_tables(self, tbl):
        # Convert the advocates nested table into multiple tables

        tbls = {
            'advocates': tbl,
            'emails': Table(),
            'phones': Table(),
            'memberships': Table(),
            'tags': Table(),
            'ids': Table(),
            'fields': Table(),
        }

        if not tbl:
            return tbls

        logger.info(f'Retrieved {tbl.num_rows} advocates...')

        # Unpack all of the single objects
        # The CapitolCanary API docs says that created_at and updated_at are dictionaries, but
        # the data returned from the server is a ISO8601 timestamp. - EHS, 05/21/2020
        for c in ['address', 'districts']:
            tbl.unpack_dict(c)

        # Unpack all of the arrays
        child_tables = [child for child in tbls.keys() if child != 'advocates']
        for c in child_tables:
            tbls[c] = tbl.long_table(['id'], c, key_rename={'id': 'advocate_id'})

        return tbls

    def get_campaigns(self, state=None, zip=None, include_generic=False, include_private=False,
                      include_content=True):
        """
        Returns a list of campaigns

        `Args:`
            state: str
                Filter by US postal abbreviation for a state or territory e.g., "CA" "NY" or "DC"
            zip: int
                Filter by 5 digit zip code
            include_generic: boolean
                When filtering by state or ZIP code, include unrestricted campaigns
            include_private: boolean
                If true, will include private campaigns in results
            include_content: boolean
                If true, include campaign content fields, which may vary. This may cause
                sync errors.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        args = {'state': state,
                'zip': zip,
                'includeGeneric': str(include_generic),
                'includePrivate': str(include_private)
                }

        tbl = Table(self.client.get_request('campaigns', params=args))
        if tbl:
            tbl.unpack_dict('updated_at')
            if include_content:
                tbl.unpack_dict('content')

        return tbl

    def create_advocate(self,
                        campaigns,
                        first_name=None,
                        last_name=None,
                        email=None,
                        phone=None,
                        address1=None,
                        address2=None,
                        city=None,
                        state=None,
                        zip5=None,
                        sms_optin=None,
                        email_optin=None,
                        sms_optout=None,
                        email_optout=None,
                        **kwargs):
        """
        Create an advocate.

        If you want to opt an advocate into or out of SMS / email campaigns, you must provide
        the email address or phone number (accordingly).

        The list of arguments only partially covers the fields that can be set on the advocate.
        For a complete list of fields that can be updated, see
        `the CapitolCanary API documentation <https://docs.phone2action.com/#calls-create>`_.

        `Args:`
            campaigns: list
                The ID(s) of campaigns to add the advocate to
            first_name: str
                `Optional`; The first name of the advocate
            last_name: str
                `Optional`; The last name of the advocate
            email: str
                `Optional`; An email address to add for the advocate. One of ``email`` or ``phone``
                is required.
            phone: str
                `Optional`; An phone # to add for the advocate. One of ``email`` or ``phone`` is
                required.
            address1: str
                `Optional`; The first line of the advocates' address
            address2: str
                `Optional`; The second line of the advocates' address
            city: str
                `Optional`; The city of the advocates address
            state: str
                `Optional`; The state of the advocates address
            zip5: str
                `Optional`; The 5 digit Zip code of the advocate
            sms_optin: boolean
                `Optional`; Whether to opt the advocate into receiving text messages; an SMS
                confirmation text message will be sent. You must provide values for the ``phone``
                and ``campaigns`` arguments.
            email_optin: boolean
                `Optional`; Whether to opt the advocate into receiving emails. You must provide
                values for the ``email`` and ``campaigns`` arguments.
            sms_optout: boolean
                `Optional`; Whether to opt the advocate out of receiving text messages. You must
                provide values for the ``phone`` and ``campaigns`` arguments. Once an advocate is
                opted out, they cannot be opted back in.
            email_optout: boolean
                `Optional`; Whether to opt the advocate out of receiving emails. You must
                provide values for the ``email`` and ``campaigns`` arguments. Once an advocate is
                opted out, they cannot be opted back in.
            **kwargs:
                Additional fields on the advocate to update
        `Returns:`
            The int ID of the created advocate.
        """

        # Validate the passed in arguments

        if not campaigns:
            raise ValueError(
                'When creating an advocate, you must specify one or more campaigns.')

        if not email and not phone:
            raise ValueError(
                'When creating an advocate, you must provide an email address or a phone number.')

        if (sms_optin or sms_optout) and not phone:
            raise ValueError(
                'When opting an advocate in or out of SMS messages, you must specify a valid '
                'phone and one or more campaigns')

        if (email_optin or email_optout) and not email:
            raise ValueError(
                'When opting an advocate in or out of email messages, you must specify a valid '
                'email address and one or more campaigns')

        # Align our arguments with the expected parameters for the API
        payload = {
            'email': email,
            'phone': phone,
            'firstname': first_name,
            'lastname': last_name,
            'address1': address1,
            'address2': address2,
            'city': city,
            'state': state,
            'zip5': zip5,
            'smsOptin': 1 if sms_optin else None,
            'emailOptin': 1 if email_optin else None,
            'smsOptout': 1 if sms_optout else None,
            'emailOptout': 1 if email_optout else None,
        }

        # Clean up any keys that have a "None" value
        payload = {
            key: val
            for key, val in payload.items()
            if val is not None
        }

        # Merge in any kwargs
        payload.update(kwargs)

        # Turn into a list of items so we can append multiple campaigns
        campaign_keys = [('campaigns[]', val) for val in campaigns]
        data = [(key, value) for key, value in payload.items()] + campaign_keys

        # Call into the CapitolCanary API
        response = self.client.post_request('advocates', data=data)
        return response['advocateid']

    def update_advocate(self,
                        advocate_id,
                        campaigns=None,
                        email=None,
                        phone=None,
                        sms_optin=None,
                        email_optin=None,
                        sms_optout=None,
                        email_optout=None,
                        **kwargs):
        """
        Update the fields of an advocate.

        If you want to opt an advocate into or out of SMS / email campaigns, you must provide
        the email address or phone number along with a list of campaigns.

        The list of arguments only partially covers the fields that can be updated on the advocate.
        For a complete list of fields that can be updated, see
        `the CapitolCanary API documentation <https://docs.phone2action.com/#calls-create>`_.

        `Args:`
            advocate_id: integer
                The ID of the advocate being updates
            campaigns: list
                `Optional`; The ID(s) of campaigns to add the user to
            email: str
                `Optional`; An email address to add for the advocate (or to use when opting in/out)
            phone: str
                `Optional`; An phone # to add for the advocate (or to use when opting in/out)
            sms_optin: boolean
                `Optional`; Whether to opt the advocate into receiving text messages; an SMS
                confirmation text message will be sent. You must provide values for the ``phone``
                and ``campaigns`` arguments.
            email_optin: boolean
                `Optional`; Whether to opt the advocate into receiving emails. You must provide
                values for the ``email`` and ``campaigns`` arguments.
            sms_optout: boolean
                `Optional`; Whether to opt the advocate out of receiving text messages. You must
                provide values for the ``phone`` and ``campaigns`` arguments. Once an advocate is
                opted out, they cannot be opted back in.
            email_optout: boolean
                `Optional`; Whether to opt the advocate out of receiving emails. You must
                provide values for the ``email`` and ``campaigns`` arguments. Once an advocate is
                opted out, they cannot be opted back in.
            **kwargs:
                Additional fields on the advocate to update
        """

        # Validate the passed in arguments
        if (sms_optin or sms_optout) and not (phone and campaigns):
            raise ValueError(
                'When opting an advocate in or out of SMS messages, you must specify a valid '
                'phone and one or more campaigns')

        if (email_optin or email_optout) and not (email and campaigns):
            raise ValueError(
                'When opting an advocate in or out of email messages, you must specify a valid '
                'email address and one or more campaigns')

        # Align our arguments with the expected parameters for the API
        payload = {
            'advocateid': advocate_id,
            'campaigns': campaigns,
            'email': email,
            'phone': phone,
            'smsOptin': 1 if sms_optin else None,
            'emailOptin': 1 if email_optin else None,
            'smsOptout': 1 if sms_optout else None,
            'emailOptout': 1 if email_optout else None,
            # remap first_name / last_name to be consistent with updated_advocates
            'firstname': kwargs.pop('first_name', None),
            'lastname': kwargs.pop('last_name', None),
        }

        # Clean up any keys that have a "None" value
        payload = {
            key: val
            for key, val in payload.items()
            if val is not None
        }

        # Merge in any kwargs
        payload.update(kwargs)

        # Turn into a list of items so we can append multiple campaigns
        campaigns = campaigns or []
        campaign_keys = [('campaigns[]', val) for val in campaigns]
        data = [(key, value) for key, value in payload.items()] + campaign_keys

        # Call into the CapitolCanary API
        self.client.post_request('advocates', data=data)



from parsons.databases.database.constants import (
    SMALLINT, MEDIUMINT, INT, BIGINT, FLOAT, BOOL, VARCHAR)

from parsons.databases.database.database import DatabaseCreateStatement

import pytest


@pytest.fixture
def dcs():
    db = DatabaseCreateStatement()
    return db


@pytest.fixture
def dcs_bool():
    db = DatabaseCreateStatement()
    db.DO_PARSE_BOOLS = True
    return db


@pytest.mark.parametrize(
    ("int1", "int2", "higher"),
    ((SMALLINT, SMALLINT, SMALLINT),
     (SMALLINT, MEDIUMINT, MEDIUMINT),
     (SMALLINT, INT, INT),
     (SMALLINT, BIGINT, BIGINT),
     (MEDIUMINT, SMALLINT, MEDIUMINT),
     (MEDIUMINT, MEDIUMINT, MEDIUMINT),
     (MEDIUMINT, INT, INT),
     (MEDIUMINT, BIGINT, BIGINT),
     (INT, SMALLINT, INT),
     (INT, MEDIUMINT, INT),
     (INT, INT, INT),
     (INT, BIGINT, BIGINT),
     (BIGINT, SMALLINT, BIGINT),
     (BIGINT, MEDIUMINT, BIGINT),
     (BIGINT, INT, BIGINT),
     (BIGINT, BIGINT, BIGINT),
     (None, BIGINT, BIGINT),
     (INT, None, INT),
     ))
def test_get_bigger_int(dcs, int1, int2, higher):
    assert dcs.get_bigger_int(int1, int2) == higher


@pytest.mark.parametrize(
    ("val", "is_valid"),
    ((10, True),
     (1_0, True),
     (+10, True),
     (+1_0, True),
     (1.2, True),
     (1.0_0, True),
     (1., True),
     (1_0., True),
     (+1.2, True),
     (+1., True),
     (+1.0_0, True),
     (0, True),
     (0.0, True),
     ("10", True),
     ("1_0", False),
     ("+10", True),
     ("+1_0", False),
     ("1.2", True),
     ("1.0_0", False),
     ("1.", True),
     ("1_0.", False),
     ("+1.2", True),
     ("+1.", True),
     ("+1.0_0", False),
     ("0", True),
     ("0.0", True),
     (True, False),
     ("True", False),
     ("a string", False),
     ({}, False),
     ([], False),
     ([], False),
     (None, False),
     ))
def test_is_valid_sql_num(dcs, val, is_valid):
    assert dcs.is_valid_sql_num(val) == is_valid


@pytest.mark.parametrize(
    ("val", "cmp_type", "detected_type"),
    ((1, None, SMALLINT),
     (1, "", SMALLINT),
     (1, MEDIUMINT, MEDIUMINT),
     (32769, None, MEDIUMINT),
     (32769, BIGINT, BIGINT),
     (2147483648, None, BIGINT),
     (2147483648, FLOAT, FLOAT),
     (5.001, None, FLOAT),
     (5.001, "", FLOAT),
     ("FALSE", VARCHAR, VARCHAR),
     ("word", "", VARCHAR),
     ("word", INT, VARCHAR),
     ("1_2", BIGINT, VARCHAR),
     ("01", FLOAT, VARCHAR),
     ("00001", None, VARCHAR),
     ("word", None, VARCHAR),
     ("1_2", None, VARCHAR),
     ("01", None, VARCHAR),
     ("{}", None, VARCHAR),
     ))
def test_detect_data_type(dcs, val, cmp_type, detected_type):
    assert dcs.detect_data_type(val, cmp_type) == detected_type


@pytest.mark.parametrize(
    ("val", "cmp_type", "detected_type"),
    ((2, None, SMALLINT),
     (2, "", SMALLINT),
     (1, MEDIUMINT, MEDIUMINT),
     (2, BOOL, SMALLINT),
     (True, None, BOOL),
     (0, None, BOOL),
     (1, None, BOOL),
     (1, BOOL, BOOL),
     ("F", None, BOOL),
     ("FALSE", None, BOOL),
     ("Yes", None, BOOL)
     ))
def test_detect_data_type_bools(dcs_bool, val, cmp_type, detected_type):
    assert dcs_bool.detect_data_type(val, cmp_type) == detected_type


@pytest.mark.parametrize(
    ("col", "renamed"),
    (("a", "a"),
     ("A", "a"),
     ("", "_"),
     ("SELECT", "select_"),
     ("two words", "two_words"),
     ("   trailing space   ", "trailing_space"),
     ("1234567890", "x_1234567890"),
     ("0word", "x_0word"),

     # create a really long column name
     # len("asdfghjkla" * 13) == 130
     # len("asdfghjkla" * 10) == 100
     ("asdfghjkla" * 13, "asdfghjkla" * 10),
     ))
def test_default_format_column(dcs, col, renamed):
    assert dcs.format_column(col) == renamed


@pytest.mark.parametrize(
    ("cols", "cols_formatted"),
    ((["a", "A", "b", "   b   ", "col name", "col_name"],
      ["a", "a_1", "b", "b_3", "col_name", "col_name_5"]),
     ))
def test_default_format_columns(dcs, cols, cols_formatted):
    assert dcs.format_columns(cols) == cols_formatted



import io
import pytest
from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart
from email.mime.image import MIMEImage
from email.mime.audio import MIMEAudio
from email.mime.application import MIMEApplication
from email.mime.base import MIMEBase

from parsons.notifications.sendmail import EmptyListError, SendMail


class TestSendMailCreateMessageSimple:
    def test_creates_mimetext_message(self):
        message = SendMail()._create_message_simple("from", "to", "subject", "text")
        assert isinstance(message, MIMEText)

    def test_message_contents_set_appropriately(self):
        message = SendMail()._create_message_simple("from", "to", "subject", "text")
        assert message.get("from") == "from"
        assert message.get("to") == "to"
        assert message.get("subject") == "subject"
        assert message.get_payload() == "text"


class TestSendMailCreateMessageHtml:
    def test_creates_multipart_message(self):
        message = SendMail()._create_message_html("from", "to", "subject", "text", "html")
        assert isinstance(message, MIMEMultipart)

    def test_sets_to_from_subject(self):
        message = SendMail()._create_message_html("from", "to", "subject", "text", "html")
        assert message.get("from") == "from"
        assert message.get("to") == "to"
        assert message.get("subject") == "subject"

    def test_works_if_no_message_text(self):
        message = SendMail()._create_message_html("from", "to", "subject", None, "html")
        assert len(message.get_payload()) == 1
        assert message.get_payload()[0].get_payload() == "html"
        assert message.get_payload()[0].get_content_type() == "text/html"

    def test_works_with_text_and_html(self):
        message = SendMail()._create_message_html("from", "to", "subject", "text", "html")
        assert len(message.get_payload()) == 2
        assert message.get_payload()[0].get_payload() == "text"
        assert message.get_payload()[0].get_content_type() == "text/plain"
        assert message.get_payload()[1].get_payload() == "html"
        assert message.get_payload()[1].get_content_type() == "text/html"


class TestSendMailCreateMessageAttachments:
    def test_creates_multipart_message(self):
        message = SendMail()._create_message_attachments("from", "to", "subject", "text", [])
        assert isinstance(message, MIMEMultipart)

    def test_can_handle_html(self):
        message = SendMail()._create_message_attachments("from", "to", "subject", "text", [],
                                                         message_html="html")
        assert len(message.get_payload()) == 2
        assert message.get_payload()[0].get_payload() == "text"
        assert message.get_payload()[0].get_content_type() == "text/plain"
        assert message.get_payload()[1].get_payload() == "html"
        assert message.get_payload()[1].get_content_type() == "text/html"

    @pytest.mark.parametrize(
        "filename,expected_type",
        [
            ("image.png", MIMEImage),
            ("application.exe", MIMEApplication),
            ("text.txt", MIMEText),
            ("audio.mp3", MIMEAudio),
            ("video.mp4", MIMEBase)  # This will fail if the method is updated to parse video
        ]
    )
    def test_properly_detects_file_types(self, tmp_path, filename, expected_type):
        filename = tmp_path / filename
        filename.write_bytes(b"Parsons")
        message = SendMail()._create_message_attachments("from", "to", "subject", "text",
                                                         [filename])
        assert len(message.get_payload()) == 2  # text body plus attachment
        assert isinstance(message.get_payload()[1], expected_type)

    @pytest.mark.parametrize("buffer", [io.StringIO, io.BytesIO])
    def test_works_with_buffers(self, buffer):
        value = "Parsons"
        if buffer is io.BytesIO:
            value = b"Parsons"
        message = SendMail()._create_message_attachments("from", "to", "subject", "text",
                                                         [buffer(value)])
        assert len(message.get_payload()) == 2  # text body plus attachment
        assert isinstance(message.get_payload()[1], MIMEApplication)


class TestSendMailValidateEmailString:
    @pytest.mark.parametrize("bad_email", ["a", "a@", "a+b", "@b.com"])
    def test_errors_with_invalid_emails(self, bad_email):
        with pytest.raises(ValueError):
            SendMail()._validate_email_string(bad_email)

    @pytest.mark.parametrize("good_email", ["a@b", "a+b@c", "a@d.com", "a@b.org"])
    def test_passes_valid_emails(self, good_email):
        SendMail()._validate_email_string(good_email)


class TestSendMailSendEmail:

    @pytest.fixture(scope="function")
    def patched_sendmail(self):
        class PatchedSendMail(SendMail):
            def _send_message(self, message):
                self.message = message  # Stores message for post-call introspection
        return PatchedSendMail()

    def test_errors_when_send_message_not_implemented(self):
        with pytest.raises(
                NotImplementedError,
                match="_send_message must be implemented for send_email to run"
        ):
            SendMail().send_email("from@from.com", "to@to.com", "subject", "text")

    def test_can_handle_lists_of_emails(self, patched_sendmail):
        patched_sendmail.send_email(
            "from",
            ["to1@to1.com", "to2@to2.com"],
            "subject", "text"
        )
        assert patched_sendmail.message.get("to") == "to1@to1.com, to2@to2.com"

    def test_errors_if_an_email_in_a_list_doesnt_validate(self, patched_sendmail):
        with pytest.raises(ValueError, match="Invalid email address"):
            patched_sendmail.send_email(
                "from",
                ["to1@to1.com", "invalid", "to2@to2.com"],
                "subject", "text"
            )

    def test_errors_if_no_to_email_is_specified(self, patched_sendmail):
        with pytest.raises(EmptyListError, match="Must contain at least 1 email"):
            patched_sendmail.send_email(
                "from",
                [],
                "subject", "text"
            )

    def test_appropriately_dispatches_html_email(self, patched_sendmail):
        patched_sendmail.send_email(
            "from", "to@to.com",
            "subject", "text", message_html="html"
        )
        assert len(patched_sendmail.message.get_payload()) == 2
        assert patched_sendmail.message.get_payload()[1].get_content_type() == "text/html"

    def test_appropriately_handles_filename_specified_as_string(self, tmp_path, patched_sendmail):
        filename = tmp_path / "test.txt"
        filename.write_bytes(b"Parsons")
        patched_sendmail.send_email(
            "from", "to@to.com",
            "subject", "text", files=str(filename)
        )
        assert len(patched_sendmail.message.get_payload()) == 2
        assert isinstance(patched_sendmail.message.get_payload()[1], MIMEText)



import unittest
import os
import requests_mock
import unittest.mock as mock
from parsons import VAN, Table
from test.utils import assert_matching_tables


class TestNGPVAN(unittest.TestCase):

    def setUp(self):

        self.van = VAN(os.environ['VAN_API_KEY'], db="MyVoters", raise_for_status=False)

    @requests_mock.Mocker()
    def test_get_changed_entity_resources(self, m):

        json = ['ActivistCodes', 'ContactHistory', 'Contacts', 'ContactsActivistCodes']
        m.get(self.van.connection.uri + 'changedEntityExportJobs/resources', json=json)
        self.assertEqual(json, self.van.get_changed_entity_resources())

    @requests_mock.Mocker()
    def test_get_changed_entity_resource_fields(self, m):

        json = [{
            'fieldName': 'ActivistCodeID',
            'fieldType': 'N',
            'maxTextboxCharacters': None,
            'isCoreField': True,
            'availableValues': None
        }, {
            'fieldName': 'ActivistCodeType',
            'fieldType': 'T',
            'maxTextboxCharacters': 20,
            'isCoreField': True,
            'availableValues': None
        }, {
            'fieldName': 'Campaign',
            'fieldType': 'T',
            'maxTextboxCharacters': 150,
            'isCoreField': True,
            'availableValues': None
        }]

        m.get(self.van.connection.uri + 'changedEntityExportJobs/fields/ActivistCodes', json=json)
        assert_matching_tables(
            Table(json), self.van.get_changed_entity_resource_fields('ActivistCodes'))

    @requests_mock.Mocker()
    def test_get_changed_entities(self, m):

        json = {"dateChangedFrom": "2021-10-10T00:00:00-04:00",
                "dateChangedTo": "2021-10-11T00:00:00-04:00",
                "files": [],
                "message": "Created export job",
                "code": None,
                "exportedRecordCount": 0,
                "exportJobId": 2170181229,
                "jobStatus": "Pending"}

        json2 = {
                    "dateChangedFrom": "2021-10-10T00:00:00-04:00",
                    "dateChangedTo": "2021-10-11T00:00:00-04:00",
                    "files": [
                        {"downloadUrl": "https://box.com/file.csv",
                         "dateExpired": "2021-11-03T15:27:01.8687339-04:00"}
                    ],
                    "message": "Finished processing export job",
                    "code": None,
                    "exportedRecordCount": 6110,
                    "exportJobId": 2170181229,
                    "jobStatus": "Complete"}

        tbl = Table([{'a': 1, 'b': 2}])

        m.post(self.van.connection.uri + 'changedEntityExportJobs', json=json)
        m.get(self.van.connection.uri + 'changedEntityExportJobs/2170181229', json=json2)

        Table.from_csv = mock.MagicMock()
        Table.from_csv.return_value = tbl

        out_tbl = self.van.get_changed_entities('ContactHistory', '2021-10-10')

        assert_matching_tables(out_tbl, tbl)



"""NGPVAN Changed Entities"""

from parsons.etl.table import Table
import logging
import time

logger = logging.getLogger(__name__)

RETRY_RATE = 10


class ChangedEntities(object):

    def __init__(self):

        pass

    def get_changed_entity_resources(self):
        """
        Get changed entity resources available to the API user.

        `Returns:`
            list
        """

        r = self.connection.get_request('changedEntityExportJobs/resources')
        logger.info(f'Found {len(r)} changed entity resources.')
        return r

    def get_changed_entity_resource_fields(self, resource_type):
        """
        Get export fields avaliable for each changed entity resource.

        `Args:`
            resource_type: str
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        tbl = Table(self.connection.get_request(f'changedEntityExportJobs/fields/{resource_type}'))
        logger.info(f'Found {tbl.num_rows} fields for {resource_type}.')
        return tbl

    def get_changed_entities(self, resource_type, date_from, date_to=None, include_inactive=False,
                             requested_fields=None, custom_fields=None):
        """
        Get modified records for VAN from up to 90 days in the past.

        `Args:`
            resource_type: str
                The type of resource to export. Use the :py:meth:`~parsons.ngpvan.changed_entities.ChangedEntities.get_changed_entity_resources`
                to get a list of potential entities.
            date_from: str
                The start date in which to search. Must be less than 90 days in the
                past. Must be``iso8601`` formatted date (``2021-10-11``).
            date_to: str
                The end date to search. Must be less than 90 days in the
                past. Must be``iso8601`` formatted date (``2021-10-11``).
            include_inactive: boolean
                Include inactive records
            requested_fields: list
                A list of optional requested fields to include. These options can be accessed through
                :py:meth:`~parsons.ngpvan.changed_entities.ChangedEntities.get_changed_entity_resource_fields`
                method.
            custom_fields: list
                A list of ids of custom fields to include in the export.

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """  # noqa: E501

        json = {
            "dateChangedFrom": date_from,
            "dateChangedTo": date_to,
            "resourceType": resource_type,
            "requestedFields": requested_fields,
            "requestedCustomFieldIds": custom_fields,
            "fileSizeKbLimit": 100000,
            "includeInactive": include_inactive
        }

        r = self.connection.post_request('changedEntityExportJobs', json=json)

        while True:
            status = self._get_changed_entity_job(r['exportJobId'])
            if status['jobStatus'] in ['Pending', 'InProcess']:
                logger.info('Waiting on export file.')
                time.sleep(RETRY_RATE)
            elif status['jobStatus'] == 'Complete':
                return Table.from_csv(status['files'][0]['downloadUrl'])
            else:
                raise ValueError(status['message'])

    def _get_changed_entity_job(self, job_id):

        r = self.connection.get_request(f'changedEntityExportJobs/{job_id}')
        return r




class Universes:
    """A class to access the Universes PDI API endpoint."""

    def __init__(self):
        universes_endpoint = "/universes"
        self.url_universes = self.base_url + universes_endpoint

        super().__init__()

    def get_universes(self, limit=None):
        """
        Get a list of Universes.

        `Args:`
            limit: int
                The number of universes to return.
        `Returns:`
            parsons.Table
                A Parsons table of all the data.
        """

        return self._request(self.url_universes, limit=limit)

    def get_universe(self, id: str):
        """
        Get a Universe by id.

        `Args:`
            id: str
                The Universe id
        `Returns:`
            parsons.Table
                A Parsons table of all the data.
        """
        return self._request(f"{self.url_universes}/{id}")



import re

__all__ = ['redact_credentials']


def redact_credentials(sql):
    """
    Redact any credentials explicitly represented in SQL (e.g. COPY statement)
    """

    pattern = "credentials\s+'(.+\n?)+[^(\\)]'"  # noqa: W605
    sql_censored = re.sub(pattern, 'CREDENTIALS REDACTED', sql, flags=re.IGNORECASE)

    return sql_censored
